"""
The summary module works similarly in logic as in the Tensorflow summary
mechanism. The API creates summary ops that would be executed later.

To say an example, to do histogram summary on certain variables::

    A.summary.histogram(name, var)

Note that the `var` needs to be named, and the `name` parameter should be the
name of the `var`. This is because not all variables are named in
`akid`. `akid` is a front-end to the underlying backend, and in backend such as
PyTorch, it does not have a unified naming mechanism. To provide unified
interface, `akid` builds its own naming mechanism. If a variable is already
named, you can call::

    A.get_name(var)

to get the name of the variable.

By default, the summary op would be put in the collection
`DEFAULT_COLLECTION`. More collections can be found at `akid.core.common`.

Currently, the summary ops are executed in a process that is parallel to the
main process.
"""
from __future__ import absolute_import
from __future__ import print_function
import multiprocessing
import six.moves.queue
import six

import numpy as np

import torch as th
from torch.autograd import Variable
from tensorboardX import SummaryWriter

from akid.utils import glog as log

from . import computational_graph as cg
from .. import computational_graph as cg_general

# Collections to keep summary ops, so it can be run later in Kid.
from akid.core.common import (
    TRAIN_SUMMARY_COLLECTION,
    VALID_SUMMARY_COLLECTION,
    TRAINING_DYNAMICS_COLLECTION,
    DEFAULT_COLLECTION,
    BATCH_MONITORING_COLLECTION
)
from akid.utils.tools import currentframe
import six
from six.moves import zip


_collections = None
_done_event = None
_queue = None
QUEUE_SIZE = 1000
summary_writer = None
_denormalize = None


class Denormalize(object):
    """De-normalize an tensor image with mean and standard deviation.

    Given mean: (R, G, B) and std: (R, G, B),
    will normalize each channel of the torch.*Tensor, i.e.
    channel = channel * std + mean

    Args:
        mean (sequence): Sequence of means for R, G, B channels respecitvely.
        std (sequence): Sequence of standard deviations for R, G, B channels
            respectively.
    """

    def __init__(self, mean, std):
        # TODO: handle the manual type cast here. Wont' work at least when in
        # cpu tensor.
        self.mean = Variable(th.Tensor(mean).cuda())
        self.std = Variable(th.Tensor(std).cuda())

    def __call__(self, tensor):
        """
        Args:
            tensor (Tensor): Tensor image of size (C, H, W) to be de-normalized.

        Returns:
            Tensor: Normalized image.
        """
        tensor = th.mul(tensor, self.std[None, None, :])
        tensor = th.add(tensor, self.mean[None, None, :])
        return tensor


def set_normalization(mean, std):
    """
    Set the normalization values for image summary, which is used to
    de-normalize the image for better view, given that the normalized images are
    hard to recognize.
    """
    global _denormalize
    _denormalize = Denormalize(mean, std)


def reset_collections():
    global _collections

    _collections = {}
    _collections[TRAIN_SUMMARY_COLLECTION] = []
    _collections[VALID_SUMMARY_COLLECTION] = []
    _collections[TRAINING_DYNAMICS_COLLECTION] = []
    _collections[DEFAULT_COLLECTION] = []
    _collections[BATCH_MONITORING_COLLECTION] = []


class SummaryOp(object):
    def __init__(self, name):
        if name is None:
            raise ValueError("Name of the summary cannot be None.")

        self.name = name
        # Save where the summary is initialized.
        f = currentframe()
        f = f.f_back.f_back.f_back
        co = f.f_code
        self.position = (co.co_filename, f.f_lineno, co.co_name)

    def __str__(self):
        return "<{} for {}> in {}".format(type(self).__name__, self.name, self.position)

    def __repr__(self):
        return "<{} for {}> in {}".format(type(self).__name__, self.name, self.position)

    def __call__(self, step, v):
        try:
            self._call(step, v)
        except Exception as e:
            print(self.__str__())
            raise e


class HistogramSummaryOp(SummaryOp):
    def __init__(self, summary_on_grad=False, **kwargs):
        super(HistogramSummaryOp, self).__init__(**kwargs)
        self.summary_on_grad = summary_on_grad

    def _call(self, step, tensor):
        # We skip illegal numbers such Nan.
        tensor_np = cg.eval(tensor)
        is_illegal = np.isnan(tensor_np).all()
        if is_illegal:
            return

        summary_writer.add_histogram(
            self.name, tensor_np, global_step=step, bins='auto')
        if self.summary_on_grad:
            grad = tensor.grad
            if grad is not None:
                summary_writer.add_histogram(
                    self.name + "_grad", cg.eval(grad), global_step=step, bins='auto')
            else:
                log.debug("Gradient of {} is None. Skip summary.".format(self.name))


class ScalarSummaryOp(SummaryOp):
    def _call(self, step, t):
        v = cg.eval(t)  if type(t) is Variable else t
        summary_writer.add_scalar(
            self.name, v, global_step=step)


class ImageSummaryOp(SummaryOp):
    def _call(self, step, t):
        if len(t.size()) == 4:
            t = t[0]
            # Since TensorboardX 1.5, it handles the torch format (CHW) by
            # default. Permuted shape without specifying data format would lead
            # to errors.
            # t = t.permute(1, 2, 0)
        if _denormalize is not None:
            t = _denormalize(t)
        if cg.torch_version < 0.4:
            v = cg.eval(t)  if type(t) is Variable else t
        else:
            v = cg.eval(t)  if type(t) is th.Tensor else t
        summary_writer.add_image(
            self.name + ' image', v, global_step=step)


def _summary_writer_worker(dir, queue, done_event):
    global summary_writer
    summary_writer = SummaryWriter(dir)
    while True:
        if done_event.is_set() and queue.empty():
            return

        try:
            op, v, step = queue.get(timeout=cg_general.TIMEOUT)
            op(step, v)
        except six.moves.queue.Empty:
            continue


def init(dir=None):
    """
    Create the summary file writer.

    It creates a process that polls a queue to see if are there anymore summary
    to write.

    Args:
        dir: str
            The directory to save event files.
    """
    global _done_event
    global _queue
    _done_event = multiprocessing.Event()
    _queue = multiprocessing.Queue(QUEUE_SIZE)
    worker_process = multiprocessing.Process(target=_summary_writer_worker, args=(dir, _queue, _done_event))
    worker_process.start()


def _add_op_to_collections(op, collections):
    if collections is None:
        collections = [DEFAULT_COLLECTION]

    for c in collections:
        _collections[c].append(op)


def histogram(name, values, summary_on_grad=False, collections=None):
    """
    Args:
        summary_on_grad: bool
            Whether to do summary on the grad of the variable.
    """
    op = HistogramSummaryOp(name=name, summary_on_grad=summary_on_grad)
    _add_op_to_collections(op, collections)


def scalar(name, value, collections=None):
    op = ScalarSummaryOp(name)
    _add_op_to_collections(op, collections)


def image(name, value, collections=None):
    op = ImageSummaryOp(name)
    _add_op_to_collections(op, collections)


def add_scalar(name, value, step):
    op = ScalarSummaryOp(name)
    _queue.put((op, value, step))


def add_graph():
    """Does nothing. Since the implementation of tensorboard-pytorch is buggy now."""
    pass


def get_collection(name=None):
    """
    If name is None, return all summary ops.
    """
    if name:
        return _collections[name]
    else:
        ret = []
        for v in six.itervalues(_collections):
            ret.extend(v)
        return ret


def merge(l):
    """Does nothing. Just to be compatible with Tensorflow backend"""
    return l


def _tensor_cleanup(v):
    """
    Detach torch tensor, and move it to cpu if necessary.
    """
    v = v.detach()
    if v.is_cuda:
        v = v.to("cpu")

    return v


# TODO: since queue is used, we do not need to wait to run such merged summary;
# we can add the summary to be written to on-the-fly during training.
def run_summary_op(ops, feed_dict=None):
    """
    Args:
        op: a list of SummaryOp.
        feed_dict: not used. Tensorflow compatibility.
    """
    if isinstance(ops, SummaryOp):
        ops = [ops]
    elif type(ops) is list:
        pass
    else:
        raise ValueError("ops should be a `SummaryOp` or a list; got {}".format(type(op)))

    # NOTE: it may take some time to transfer from GPU to CPU, if this becomes
    # a bottleneck, a thread should be used to transfer the data.
    op_values = []
    for op in ops:
        t = cg.tensor_by_name[op.name]
        if isinstance(t, th.Tensor):
            op_values.append(_tensor_cleanup(t))
        elif isinstance(t, cg_general.NamedTensorTuple):
            op_values.append([_tensor_cleanup(v) for v in t])
        elif isinstance(t, cg_general.NamedScalar) or isinstance(t, cg_general.NamedNdarrary):
            op_values.append(t)
        else:
            raise ValueError("{} is not a valid value type".format(type(t)))

    op_value_tuples = list(zip(ops, op_values))
    for t in op_value_tuples:
        try:
            # TODO: resize queue if its size is not enough.
            _queue.put((t[0], t[1], cg_general.get_step()))
        except RuntimeError as e:
            # TODO: The runtime error cannot be caught. Possibly because another
            # thread actually is doing this other than the main process.
            six.raise_from(ValueError("Error when putting {} into queue.".format(t[0])), e)


def close():
    if _done_event is not None:
        _done_event.set()


reset_collections()



from __future__ import absolute_import
from sklearn.metrics import roc_auc_score, average_precision_score
import numpy as np

from ..core.blocks import ProcessingLayer
from ..backend import NamedTensorTuple, NamedScalar
from .. import backend as A
from .. import nn
from six.moves import range


class EvalLayer(ProcessingLayer):
    NAME = "Eval"

    @property
    def data(self):
        return self._eval


class AUCLayer(EvalLayer):
    """
    Args:
        return_average_auc: bool
            If is set `True`, and the labels are multi-label, an label-wise
            averaged result would be returned. Otherwise, return a list of auc
            values, each of which (in the list) is the auc for that particular
            class.
    """
    NAME = "AUC"

    def __init__(self, return_average_auc=True, *args, **kwargs):
        super(AUCLayer, self).__init__(*args, **kwargs)

        self.return_average_auc = return_average_auc

    def _forward(self, inputs):
        v = A.eval(inputs)
        y, y_pred = v[1], v[0]

        self._eval = None
        if self.return_average_auc:
            auc = []
            for i in range(y.shape[1]):
                if np.sum(y[:, i]) != 0:
                    auc.append(roc_auc_score(y[:, i], y_pred[:, i], average='macro'))

            self._eval = NamedScalar("batch_AUC", np.mean(auc))

        if self.is_val:
            self._verbose_eval = NamedTensorTuple("AUC", (y_pred, y))

        return self._eval


class MultiLabelAccuracy(EvalLayer):
    """
    Given a tuple of (prediction score, label), compute the mean accuracy of
    multi-label classification.
    """
    def _forward(self, inputs):
        inputs = A.eval(inputs)
        for i, v in enumerate(inputs):
            shape = A.get_shape(v)
            if len(shape) != 2:
                raise ValueError("Input {} should be a 2-dim array. Got {} instead.".format(i, shape))

        mean_precision = []
        for i in range(shape[1]):
            if np.sum(inputs[1][:, i]) != 0:
                mean_precision.append(average_precision_score(inputs[1][:, i], inputs[0][:, i]))

        mean_precision = np.mean(mean_precision)
        mean_precision = A.NamedScalar("MAP", mean_precision)
        self._eval = mean_precision

        return self._eval


class BinaryAccuracy(EvalLayer):
    NAME = "Binary_Acc"

    def __init__(self, hinge_loss_label=False, *args, **kwargs):
        super(BinaryAccuracy, self).__init__(*args, **kwargs)
        self.hinge_loss_label = hinge_loss_label

    def _forward(self, inputs):
        x, y = inputs[0], inputs[1]
        self._eval = nn.binary_accuracy(x, y, self.hinge_loss_label, name="Acc")
        return self._eval



"""
A module to provide a mechanism to ease network tuning.
"""
from __future__ import print_function

from __future__ import absolute_import
import sys
import inspect
import multiprocessing
from .semaphore import Semaphore
import subprocess
import os

from jinja2 import Template
from tqdm import tqdm
import pycuda.autoinit
import pycuda.driver as cuda
import gflags as flags
from six.moves import range

FLAGS = flags.FLAGS
flags.DEFINE_boolean("use_sub_shell", True, "Deprecated. Not used."
                     " Use sub shell to run training"
                     " instances or not. This is to get around the issue"
                     " tensorflow will not release memory after a process"
                     " finishes its work. So to force tensorflow release "
                     " resources, just run the process in a sub shell.")
flags.DEFINE_integer("gpu_start_No", 0, "The start No of GPU to use. This flag"
                     " is to make sure the correct GPU mask is passed when"
                     " running training instances using subshell. For example,"
                     " if you are going to use GPU 1-9, you need to pass"
                     " --gpu_start_no=1.")
flags.DEFINE_boolean("chk_point", False, "Whether to continue from check point."
                     " This option is for the interrupted training instances. Make sure"
                     " `log_dir` is the same to let akid find the checkpoint.")

NETWORK_LOG_HEADER = "Network Setup: \\n"


def spawn_using_sub_shell(setup_func, work_dir, idxs):
    gpu_No_str = ""
    # Make comma separated gpu No list.
    for idx in idxs:
        gpu_No_str += "{},".format(idx + FLAGS.gpu_start_No)

    # Add training code to the end.
    training_call = """


kid = setup()
import inspect
from akid.utils import glog as log
log.info("{}" + inspect.getsource(setup))
kid.practice(continue_from_chk_point={})
    """.format(NETWORK_LOG_HEADER, FLAGS.chk_point)

    training_code = setup_func + training_call
    # Save code to file.
    file_name = "net_{}.py".format(gpu_No_str)
    with open(os.path.join(work_dir, file_name), 'w') as f:
        f.write(training_code)
    # Run.
    subprocess.call(
        "cd {}; CUDA_VISIBLE_DEVICES={} python {}".format(
            work_dir,
            gpu_No_str,
            file_name),
        shell=True)


def spawn(s, l, gpu_mask, gpu_num, return_values, setup_func, repeat):
    s.acquire(gpu_num)
    print ("Acquired resources. Remaining semaphore {}".format(s.semaphore))

    with l:
        # Look up GPU(s) and mark it used.
        # The lock is to make sure the logging is not mixed.
        print ("Entering critical region.")
        acquired_gpu = 0
        idxs = []
        for idx, avail in enumerate(gpu_mask):
            if avail == 1:
                gpu_mask[idx] = 0
                idxs.append(idx)
                acquired_gpu += 1
                if acquired_gpu == gpu_num:
                    break

        print("GPU mask {}.".format(gpu_mask))
        print("Using GPU {}.".format(idxs))

        print ("Exiting critical region.")

    repeat_folder = str(repeat)
    # Create folder to hold one training repeat.
    if not os.path.exists(repeat_folder):
        os.mkdir(repeat_folder)
    work_dir = repeat_folder

    spawn_using_sub_shell(setup_func, work_dir, idxs)

    # Release the GPU.
    with l:
        print ("Entering critical region.")
        for idx in idxs:
            print("Released GPU {}.".format(idx))
            gpu_mask[idx] = 1
            print("GPU mask {}.".format(gpu_mask))

        print ("Exiting critical region.")

    s.release(gpu_num)
    print ("Released resources. Remaining semaphore {}".format(s.semaphore))

def tune(template,
         opt_paras_list=[{}],
         net_paras_list=[{}],
         repeat_times=1,
         gpu_num_per_instance=1,
         debug=False):
    """
    A function `tune` that takes a Brain jinja2 template class and a parameters
    to fill the template in runtime. Parameters provided should complete the
    remaining network parameters in the template. The tuner is not aware of the
    content of the list items. It is up to the user to define template right,
    so parameters will be filled in the right place.

    The jinja2 template must be a function named `setup`, and return a set up
    `Kid`. All necessary module imports should be put in the function instead
    of module level import usually.

    The `tune` function would use all available GPUs to train networks with all
    given different set of parameters. If available GPUs are not enough, the
    ones that cannot be trained will wait till some others finish, and get its
    turn.

    ## Parameter Tuning Usage

    Tunable parameters are divided into two set, network hyper parameters,
    `net_paras_list`, and optimization hyper parameters, `opt_paras_list`. Each
    set is specified by a list whose item is a dictionary that holds the actual
    value of whatever hyper parameters defined as jinja2 templates. Each item
    in the list corresponds to a tentative training instance. network paras and
    optimization paras combine with each other exponentially(or in Cartesian
    Product way if we could use Math terminology), which is to say if you have
    two items in network parameter list, and two in optimization parameters,
    the total number of training instances will be four.

    Final training precisions will be returned as a list. Since the final
    precision normally will not be the optimal one, which normally occurs
    during training, the returned values are used for testing purpose only now

    ## Run repeated experiment

    To run repeated experiment, just leave `opt_paras_list` and
    `net_paras_list` to their default value.

    ## GPU Resources Allocation

    If the `gpu_num_per_instance` is None, a gpu would be allocated to each
    thread, otherwise, the length of the list should be the same with that of
    the training instance (aka the #opt_paras_list * #net_paras_list *
    repeat_times), or an int.

    Given the available GPU numbers, a semaphore is created to control access
    to GPUs. A lock is created to control access to the mask to indicator which
    GPU is available. After a process has modified the gpu mask, it releases
    the lock immediately, so other process could access it. But the semaphore
    is still not release, since it is used to control access to the actual
    GPU. A training instance will be launched in a subshell using the GPU
    acquired. The semaphore is only released after the training has finished.

    ## Example

    For example, to tune the activation function and learning rates of a
    network, first we set up network parameters in `net_paras_list`,
    optimization parameters in `opt_paras_list`, build a network in the `setup`
    function, then pass all of it to tune::

        net_paras_list = []
        net_paras_list.append({
            "activation": [
                {"type": "relu"},
                {"type": "relu"},
                {"type": "relu"},
                {"type": "relu"}],
            "bn": True})
        net_paras_list.append({
            "activation": [
                {"type": "maxout", "group_size": 2},
                {"type": "maxout", "group_size": 2},
                {"type": "maxout", "group_size": 2},
                {"type": "maxout", "group_size": 5}],
            "bn": True})

        opt_paras_list = []
        opt_paras_list.append({"lr": 0.025})
        opt_paras_list.append({"lr": 0.05})

        def setup(graph):

            brain.attach(cnn_block(
                ksize=[8, 8],
                init_para={
                    "name": "uniform",
                    "range": 0.005},
                wd={"type": "l2", "scale": 0.0005},
                out_channel_num=384,
                pool_size=[4, 4],
                pool_stride=[2, 2],
                activation={{ net_paras["activation"][1] }},
                keep_prob=0.5,
                bn={{ net_paras["bn"] }}))

        tune(setup, opt_paras_list, net_paras_list)
    """
    # Parse command line flags
    FLAGS(sys.argv)
    # Set up data structures.
    # #########################################################################
    manager = multiprocessing.Manager()
    gpu_num = cuda.Device.count()
    gpu_mask = manager.list([1] * gpu_num)
    return_values = manager.list()

    net_num = len(net_paras_list)
    opt_num = len(opt_paras_list)

    if type(gpu_num_per_instance) is not int:
        if len(net_paras_list) * len(opt_paras_list) * repeat_times \
           != len(gpu_num_per_instance):
            raise Exception("""
            The number of gpu used per training instance should match
            `#net_paras_list({}) * #opt_paras_list({}) * repeat_times({}): {}`,
            or a single int.
            """.format(net_num,
                       opt_num,
                       repeat_times,
                       net_num * opt_num * repeat_times)
            )

    # Logistics
    # #########################################################################
    s = Semaphore(len(gpu_mask))
    l = multiprocessing.Lock()
    process_pool = []
    template_str = Template(inspect.getsource(template))

    # Start tuning.
    # #########################################################################
    for repeat in range(0, repeat_times):
        for i, opt_paras in enumerate(opt_paras_list):
            for j, net_paras in enumerate(net_paras_list):
                setup_func = template_str.render(opt_paras=opt_paras,
                                                 net_paras=net_paras)
                _gpu_num = gpu_num_per_instance[
                    repeat*(net_num * opt_num) + i*net_num + j] \
                    if type(gpu_num_per_instance) is list \
                    else gpu_num_per_instance
                p = multiprocessing.Process(target=spawn,
                                            args=(s,
                                                  l,
                                                  gpu_mask,
                                                  _gpu_num,
                                                  return_values,
                                                  setup_func,
                                                  repeat))
                process_pool.append(p)
                p.start()

    # Wait for all processes to finish.
    for p in tqdm(process_pool):
        p.join()

    # TODO(Shuai): Think what should be the return value for subprocess call.
    return return_values



"""
Akid aims to focus on enforcing the **way** to **build** entities/blocks that
process signal, instead of letting how signals/tensors propagates influence how
blocks are built. In this sense, the underlying tensor processing engine is
arbitrary, and abstracted as backends.

## Variable allocation

`akid` only distinguishes between variables that are trainable, and that are
not. To obtained a variable that is trainable, call `get_variable`; one that is
not, call `Tensor`. However, if any operations that involve both trainable and
non-trainable variables, the non-trainable should be declared with
`requires_grad` true, to indicate though they are not trainable, they should
save gradients (for back propagation).
"""
from __future__ import absolute_import
import os
import sys

import tensorflow as tf
import torch as th

from .computational_graph import *
from ..utils import glog
import gflags
FLAGS = gflags.FLAGS


if 'AKID_BACKEND' in os.environ:
    _backend = os.environ['AKID_BACKEND']
    assert _backend in {TF, TORCH}
    _BACKEND = _backend
else:
    _BACKEND = TF

if _BACKEND ==  TF:
    from .tf_backend import *
elif _BACKEND == TORCH:
    from .th_backend import *
else:
    raise ValueError("Backend {} is not supported.".format(_BACKEND))


def backend():
    """
    Public interface to get the backend in use.
    """
    return _BACKEND


def reset():
    reset_step()
    reset_block_count()
    close()
    # We use scope management from tensorflow, so reset graph
    tf.reset_default_graph()
    if _BACKEND == TORCH:
        summary.reset_collections()
        summary.close()


_log_initialized = False
gflags.DEFINE_string('f', '', 'kernel')

def init_log():
    global _log_initialized
    if not _log_initialized:
        FLAGS(sys.argv)
        glog.init("stdout", akid_logger=True)
        _log_initialized = True


def is_numerical(d):
    if isinstance(d, NamedNumericValue):
        return True

    if _BACKEND == TF:
        if isinstance(d, tf.Tensor):
            return True
    elif _BACKEND == TORCH:
        if isinstance(d, th.Tensor):
            return True

    return False



"""
This module offer a top level class for testing.
"""
from __future__ import absolute_import
import sys
import os
import pdb
import functools
import traceback

import unittest
from unittest import TestCase
from unittest import skip

from akid import AKID_DATA_PATH
from akid import (
    MNISTFeedSource,
    Cifar10TFSource,
    MNISTSource,
    OldSource,
    Kid,
    FeedSensor,
    SimpleSensor,
    ParallelSensor,
    MomentumKongFu
)
from akid.models.brains import OneLayerBrain
from akid import backend as A
from six.moves import map

from .debug import debug_on


def skipUnless(cond, msg=None):
    return unittest.skipUnless(cond, msg)


def main():
    unittest.main()


class AKidTestCase(TestCase):
    def setUp(self):
        A.reset()
        A.init_log()

    def tearDown(self):
        A.reset()

    def assertNdarrayEquals(self, a, b):
        if not (a == b).all():
            msg = self._formatMessage(None, '\n{}\n == \n{}\n'.format(a, b))
            raise self.failureException(msg)

    def assertNdarrayEqualAtLeastOne(self, a, b):
        for v in b:
            if (a == v).all():
                return

        msg = self._formatMessage(None, '\n{}\n == \n{}\n'.format(a, b))
        raise self.failureException(msg)

    def assertNdarrayAlmostEquals(self, a, b, places=7):
        diff = a - b
        diff = abs(diff)
        if not (diff < 10 ** -places).all():
            msg = self._formatMessage(None, '\ndiff:\n{}\n\n\n a:{}\n\n\n b:{}\n'.format(diff, a, b))
            raise self.failureException(msg)

    def assertNdarrayAlmostEqualAtLeastOne(self, a, b, places=7):
        for v in b:
            diff = a - v
            diff = abs(diff)
            if (diff < 10 ** -places).all():
                return

        msg = self._formatMessage(None, '\n{}\n == \n{}\n'.format(a, b))
        raise self.failureException(msg)

    def assertNdarrayNotAlmostEquals(self, a, b):
        diff = a - b
        diff = abs(diff)
        if (diff < 10e-7).all():
            msg = self._formatMessage(None, '\ndiff:\n{}\n a:{}\n b:\n'.format(diff, a, b))
            raise self.failureException(msg)

    def assertTensorEquals(self, a, b):
        a = A.eval(a)
        b = A.eval(b)
        if not (a == b).all():
            msg = self._formatMessage(None, '\n{}\n == \n{}\n'.format(a, b))
            raise self.failureException(msg)

    def assertTensorAlmostEquals(self, a, b, places=7):
        diff = a - b
        diff = A.eval(diff)
        diff = abs(diff)
        if not (diff < 10 ** -places).all():
            msg = self._formatMessage(None, '\ndiff:\n{}\n'.format(diff))
            raise self.failureException(msg)

    def assertTensorNotAlmostEquals(self, a, b, places=7):
        diff = a - b
        diff = A.eval(diff)
        diff = abs(diff)
        if (diff < 10 ** -places).all():
            msg = self._formatMessage(None, '\ndiff:\n{}\n'.format(diff))
            raise self.failureException(msg)


class TestFactory(object):
    """
    A top level class for testing `akid`.
    """

    @staticmethod
    def get_test_brain(using_moving_average=False):
        """
        Args:
            using_moving_average: A Boolean. Using moving average to do test or
                not for the test brain.
        """
        if using_moving_average:
            return OneLayerBrain(moving_average_decay=0.9, name="test_brain")
        else:
            return OneLayerBrain(name="test_brain")

    @staticmethod
    def get_test_feed_source():
        return MNISTFeedSource(
            name="MNIST_feed",
            url='http://yann.lecun.com/exdb/mnist/',
            num_train=60000,
            num_val=10000,
            scale=True)

    @staticmethod
    def get_test_tf_source():
        return Cifar10TFSource(
            name="CIFAR10",
            url='http://www.cs.toronto.edu/~kriz/cifar-10-binary.tar.gz',
            work_dir=AKID_DATA_PATH + '/cifar10',
            num_train=50000,
            num_val=10000)

    @staticmethod
    def get_test_sensor(simple=False):
        if A.backend() == A.TF:
            return FeedSensor(source_in=TestFactory.get_test_feed_source(),
                              val_batch_size=100,
                              name='data')
        elif A.backend() == A.TORCH:
            s = MNISTSource(work_dir=AKID_DATA_PATH + '/mnist', name='mnist')
            s.setup()
            if simple:
                return SimpleSensor(
                    source_in=s,
                    # Do not shuffle training set for reproducible test
                    sampler="sequence",
                    name='mnist')
            else:
                return ParallelSensor(
                    source_in=s,
                    # Do not shuffle training set for reproducible test
                    sampler="sequence",
                    name='mnist')


    @staticmethod
    def get_test_kid(data_in, brain):
        """
        Return a default kid given a source and a brain.
        """
        if issubclass(type(data_in), OldSource):
            return Kid(
                FeedSensor(source_in=data_in, name='data'),
                brain,
                MomentumKongFu(),
                # debug=True,
                max_steps=900)
        else:
            # data_in is a sensor now
            return Kid(
                data_in,
                brain,
                MomentumKongFu(),
                # debug=True,
                max_steps=900)


class TestSuite():
    def suite(self): #Function stores all the modules to be tested
        modules_to_test = []
        test_dir = os.listdir('.')
        for test in test_dir:
            if test.startswith('test') and test.endswith('.py'):
                modules_to_test.append(test.rstrip('.py'))
        modules_to_test.remove('test_engines')
        modules_to_test.remove('test_logging')
        modules_to_test.remove('test_sources')
        modules_to_test.remove('test_observer')

        all_tests = unittest.TestSuite()
        for module in map(__import__, modules_to_test):
            all_tests.addTest(unittest.findTestCases(module))
        return all_tests


class ForkablePdb(pdb.Pdb):

    _original_stdin_fd = sys.stdin.fileno()
    _original_stdin = None

    def __init__(self):
        pdb.Pdb.__init__(self)

    def _cmdloop(self):
        current_stdin = sys.stdin
        try:
            if not self._original_stdin:
                self._original_stdin = os.fdopen(self._original_stdin_fd)
            sys.stdin = self._original_stdin
            self.cmdloop()
        finally:
            sys.stdin = current_stdin



from __future__ import absolute_import
import inspect

import tensorflow as tf

from ..core.blocks import ProcessingLayer
from .activation_layers import GroupSoftmaxLayer
from .. import backend as A
from .. import nn
from six.moves import range


class LossLayer(ProcessingLayer):
    """
    An abstract top level loss layer.

    It is nothing here now. However, it has already been used to check whether
    a layer is a loss layer or not.
    """
    def __init__(self, multiplier=1, **kwargs):
        super(LossLayer, self).__init__(**kwargs)
        self.multiplier = multiplier

    def _setup(self):
        self.log("Using multiplier {}".format(self.multiplier))

    def _post_forward(self, *args, **kwargs):
        self._loss = A.mul(self._loss, self.multiplier, name="loss")

        super(LossLayer, self)._post_forward(self, *args, **kwargs)

        return self._loss


class SoftmaxWithLossLayer(LossLayer):
    """
    This layer supports two types of labels --- scalar label, or dense vector
    labels. Passing either one type of them will work out of the box.
    """
    def __init__(self, class_num, **kwargs):
        super(SoftmaxWithLossLayer, self).__init__(**kwargs)
        self.class_num = class_num

    def _forward(self, data_in):
        logits = data_in[0]
        labels = data_in[1]

        self._loss = A.nn.cross_entropy_loss(logits, labels, name="xentropy_mean")
        self._eval = A.nn.class_acccuracy(logits, labels, name='accuracy')
        self._data = self._loss

        return self._data


class GroupSoftmaxWithLossLayer(SoftmaxWithLossLayer, GroupSoftmaxLayer):
    """
    A loss layer that makes use of the group structure in the target classes.
    Group structure here means hidden units in a group is mutually exclusive
    classes, for instance, sub-classes in a super class, not the mathematical
    group, yet.

    The input labels should be processed as a multi-hot vector augmented for
    non-existence labels in groups (it is not convenient to fine granularity
    processing of labels with tensors). For example, suppose overall there are
    4 classes, each consecutive 2 classes make up a group. A label 3 (starting
    with 0) corresponds to a vector `100001`, where the first 2 stands for
    non-existence of the first super class (so are two sub-classes in the
    group).
    """
    NAME = "GSMax_Loss"

    def __init__(self, augment_label=False, **kwargs):
        """
        Args:
            augment_label: bool
                Augment label with non-existence label or not.
        """
        super(GroupSoftmaxWithLossLayer, self).__init__(**kwargs)
        self.augment_label = augment_label

    def _forward(self, data_in):
        input = data_in[0]
        labels = data_in[1]
        label_vectors = data_in[2]

        # Compute group softmax.
        shape = input.get_shape().as_list()
        assert len(shape) is 2, "Input should be rank 2."
        if shape[-1] % self.group_size is not 0:
            raise Exception("Group size {} should evenly divide output channel"
                      " number {}".format(self.group_size, shape[-1]))

        out_channel_num = shape[-1]
        num_split = out_channel_num // self.group_size
        self.log("Feature maps of layer {} is divided into {} group".format(
            self.name, num_split))
        data = tf.reshape(input, [-1, shape[-1]])
        if self.use_temperature:
            T = self._get_variable(
                "T",
                [1],
                initializer=tf.constant_initializer(10.0))
            data /= T
        data_split = tf.split(axis=1, num_or_size_splits=num_split, value=data)
        data_split = list(data_split)
        # Augment each split with a constant 1.
        # Get dim of non-channel shape
        nc_shape = 1
        for d in shape[0:-1]:
            nc_shape *= d
        self.ground_state = tf.constant(1.0, shape=[nc_shape, 1])
        for i in range(0, len(data_split)):
            data_split[i] = tf.concat(axis=1,
                                      values=[data_split[i],
                                       self.ground_state])
        for i in range(0, len(data_split)):
            data_split[i] = tf.nn.softmax(
                data_split[i])

        if self.augment_label:
            # All labels eval graph.
            # Note we need access to hidden units before augmented dimensions
            # for non-existence have been dropped, so this eval graph
            # constructor has not been put in the end.
            group_label_vectors = tf.split(axis=1, num_or_size_splits=num_split, value=label_vectors)
            group_label_vectors = list(group_label_vectors)

            group_label_argmax_idx = []
            for group_label_vector in group_label_vectors:
                group_label_argmax_idx.append(tf.argmax(group_label_vector, 1))
            label_argmax_idx = tf.stack(group_label_argmax_idx)

            group_data_argmax_idx = []
            for group_data_vector in data_split:
                group_data_argmax_idx.append(tf.argmax(group_data_vector, 1))
            data_argmax_idx = tf.stack(group_data_argmax_idx)

            all_label_eval = tf.reduce_mean(
                tf.cast(tf.equal(data_argmax_idx, label_argmax_idx),
                        tf.float32),
                name="augmented_acc")

            # Compute cross entropy with labels.
            # I could choose to split label vector and do cross entropy one by
            # one, or merge the splitted probability vectors and do cross
            # entropy once. The latter was chosen.
            data = tf.concat(axis=1, values=data_split,)
            aug_shape = list(shape)
            aug_shape[-1] = (self.group_size + 1) * num_split
            logits = tf.reshape(data,
                                aug_shape,
                                GroupSoftmaxWithLossLayer.NAME)
            _ = - tf.cast(label_vectors, tf.float32) * tf.log(logits)
            cross_entropy_mean = tf.reduce_mean(_, name='xentropy_mean')

        # Drop the augmented dimension.
        for i in range(0, len(data_split)):
            data_split[i] = data_split[i][:, 0:self.group_size]
        data = tf.concat(axis=1, values=data_split,)
        output = tf.reshape(data, shape, GroupSoftmaxWithLossLayer.NAME)

        self._data = output

        if not self.augment_label:
            batch_size = tf.size(labels)
            _labels = tf.expand_dims(labels, 1)
            indices = tf.expand_dims(tf.range(0, batch_size, 1), 1)
            concated = tf.concat(axis=1, values=[indices, _labels])
            onehot_labels = tf.sparse_to_dense(
                concated, tf.stack([batch_size, self.class_num]), 1.0, 0.0)
            cross_entropy \
                = tf.nn.softmax_cross_entropy_with_logits(logits=output,
                                                          labels=onehot_labels,
                                                          name='xentropy')
            cross_entropy_mean = tf.reduce_mean(cross_entropy,
                                                name='xentropy_mean')

        self._loss = cross_entropy_mean

        # Real label eval graph.
        # We check the dimension of the real label has the largest value
        # (probability).
        correct = tf.nn.in_top_k(output, labels, 1)
        # Return the number of true entries.
        real_label_eval = tf.reduce_mean(
            tf.cast(correct, tf.float32),
            name="real_acc")

        if self.augment_label:
            # Since we have augmented labels, two eval graph have been created,
            # one for the real label, and one for real and augmented labels. A
            # list will be saved to `_eval`.
            self._eval = [all_label_eval, real_label_eval]
        else:
            self._eval = real_label_eval


class MSELossLayer(LossLayer):
    NAME = "MSELoss"

    def __init__(self, size_average=True, **kwargs):
        super(MSELossLayer, self).__init__(**kwargs)

        self.size_average = size_average

    def _forward(self, data):
        self._loss = A.nn.mse_loss(data[0], data[1], self.size_average, name="mse_loss")
        self._data = self._loss
        return self._loss


class BCELossLayer(LossLayer):
    NAME = "BCELoss"

    def _forward(self, x):
        if len(x) == 3:
            pos_weight = x[2]
        else:
            pos_weight = None

        self._loss = A.nn.binary_cross_entropy_loss_with_logits(x[0], x[1], pos_weight=pos_weight)
        self._data = self._loss
        return self._loss


class WeightDecayLayer(LossLayer):
    NAME = "WD"

    def _forward(self, variable_list):
        sum = 0
        for v in variable_list:
            v = A.reshape(v, -1)
            v_square = v.dot(v)
            sum += v_square

        sum *= self.multiplier

        self._loss = sum
        self._data = sum

        return sum


class HingeLossLayer(LossLayer):
    """
    The hinge loss takes class score, and labels and compute hinge loss.

    Args:
        If the inputs is a two-element list of `[y, t]`, where `y` is the
        estimated class score, and `t` is the target label. `t` should give
        labels as +1, or -1. Optionally, a third element could be added to be
        use as weights for each sample and each label (if it is a 2D tesnor),
        or just for each sample (if it is a 1D tensor).
    """
    NAME = "HingeLoss"
    def __init__(self, tripod_hinge_loss=False, signal_balancing=False, *args, **kwargs):
        super(HingeLossLayer, self).__init__(*args, **kwargs)

        self.tripod_hinge_loss = tripod_hinge_loss
        self.signal_balancing = signal_balancing

    def _forward(self, x):
        if len(x) == 3:
            weights = x[2]
        else:
            weights = None

        if self.tripod_hinge_loss:
            self._loss = nn.tripod_hinge_loss(x[0], x[1], weights=weights, signal_balancing=self.signal_balancing, name="tripod_hinge_loss")
        else:
            self._loss = nn.hinge_loss(x[0], x[1], weights=weights, name="hinge_loss")
        return self._loss

HingeLoss = HingeLossLayer


class MarginRankingLossLayer(LossLayer):
    """
    See :meth:`akid.nn.losses.hinge_loss` for hinge ranking loss. Here we note
    the extra work done by the layer wrapper.

    In some cases, e.g., implicit matrix completion, the negative samples could
    be missing data. So it makes the model biased if we simply treat the
    missing data as negative samples. In this case, it might be appropriate to
    take an average score of negative sample. It at least gives a check on the
    bias, since of very low probability out of the randomly sampled negative
    samples, most of them are actually missed positive samples.
    """
    NAME = "MarginRankingLoss"

    def __init__(self, margin, hard_sample_mining=False, *args, **kwargs):
        super(MarginRankingLossLayer, self).__init__(*args, **kwargs)
        self.margin = margin
        self.hard_sample_mining = hard_sample_mining

    def _forward(self, x):
        positive_samples, negative_samples = x[0], x[1]

        if not isinstance(positive_samples, tuple):
            assert not isinstance(negative_samples, tuple), "If tuples are passed in, both of them should be tuples."
            # First, we check the shape of x,
            shape = A.get_shape(negative_samples)
            # NOTE: the below case leaves the case where the dims of positive
            # and negative samples are aligned, and we want to compute the
            # element-wise margin loss.
            if len(shape) == 3:
                # Means we randomly sample some negative samples, and intend to use
                # their average.
                negative_samples = A.mean(negative_samples, -1)
            elif len(shape) == 1:
                # We want to compute the margin between each positive and the
                # rest of negative samples.
                negative_samples = A.expand_dims(negative_samples, 0)
                positive_samples = A.expand_dims(positive_samples, 1)

            self._loss = nn.hinge_ranking_loss(positive_samples,
                                               negative_samples,
                                               self.margin,
                                               hard_sample_mining=self.hard_sample_mining,
                                               name="hinge_ranking_loss")
        else:
            assert isinstance(negative_samples, tuple), "If tuples are passed in, both of them should be tuples."
            # NOTE: we could sample some negative samples, instead use all
            # samples for now.
            losses = []
            for i,v in enumerate(positive_samples):
                if len(negative_samples[i].shape) == 1:
                    ns = A.expand_dims(negative_samples[i], 0)
                else:
                    ns = negative_samples[i]
                loss = nn.hinge_ranking_loss(A.expand_dims(v, 1),
                                             ns,
                                             self.margin,
                                             hard_sample_mining=self.hard_sample_mining)
                losses.append(loss)

            losses = A.stack(losses)

            self._loss = A.mean(losses, name="hinge_ranking_loss")

        return self._loss


MarginRankingLoss = MarginRankingLossLayer


class KLRegularizationLayer(LossLayer):
    """
    Intake a tensor, and compute point-wise KL divergence between empirical
    probability distribution and Gaussian distribution.
    """
    def __init__(self, mean=0, std=1, *args, **kwargs):
        super(KLRegularizationLayer, self).__init__(*args, **kwargs)

        # Mean is not used yet.
        self.mu_0 = mean
        self.sigma_0 = std

    def _forward(self, x):
        # mu = A.mean(x)
        sigma = A.std(x, dim=1)

        # Regularize to N(0, 1)
        # KL_loss = A.mul(- 0.5, 1 + A.log(sigma ** 2) - mu ** 2 - sigma ** 2, name="kl_loss")

        # Only regularize on the std.
        kl_user_loss = -0.5 + A.log(self.sigma_0/sigma) + 0.5 * (sigma / self.sigma_0)**2
        self._loss = A.mean(kl_user_loss)

        return self._loss



KLRegularization = KLRegularizationLayer


__all__ = [name for name, x in locals().items() if
           not inspect.ismodule(x) and not inspect.isabstract(x)]



from __future__ import absolute_import
import inspect

import tensorflow as tf

from ..core.blocks import ProcessingLayer
from ..core.jokers import Joker
from .. import backend as A


class ReshapeLayer(ProcessingLayer):
    """
    Reshape data. Only intrinsic shape information of output data should be
    given. The dimension of batch size is not needed. If no shape is given, all
    dimensions beyond batch size dimension are flattened. Unless `squeeze` is
    set to False, shape of dimension 1 would be squeezed.
    """
    NAME = "Reshape"

    def __init__(self, shape=None, squeeze=True, **kwargs):
        # Do not do summary since we just reshape the data.
        kwargs["do_summary"] = False
        super(ReshapeLayer, self).__init__(**kwargs)
        self.intrinsic_shape = shape
        self.squeeze = squeeze

    def _forward(self, input):
        batch_size = A.get_shape(input)[0]
        if self.intrinsic_shape:
            shape = list(self.intrinsic_shape)
        else:
            shape = [-1]
        shape.insert(0, batch_size)
        self._data = A.reshape(input, shape)

        if self.squeeze:
            self._data = A.squeeze(self._data)

        return self._data


class PaddingLayer(ProcessingLayer, Joker):
    """
    Zero padding on height and width dimensions of the input feature map.

    This layer can work with input shape [H, W, C] and [N, H, W, C].
    """
    def __init__(self, padding=[1, 1], **kwargs):
        """
        Args:
            padding: a two-element list or a list whose list is the same as the
                     input data.
                [H, W]. H holds how many padding will be added in height;
                similar W is for width. Padding is added symmetrically for each
                dimension. If not a two-element list, a full length list is
                needed.
        """
        # Do not do summary since we just pad the data.
        kwargs["do_summary"] = False
        if "name" not in kwargs:
            kwargs["name"] = "padding"
        super(PaddingLayer, self).__init__(**kwargs)
        self.padding = padding

    def _forward(self, input):
        shape = A.get_shape(input)
        assert len(shape) is 4 or 3,\
            "Shapes other than 4 or 3 are not supported."

        if len(self.padding) is 2:
            if len(shape) is 4:
                _padding = [
                    [0, 0],
                    [self.padding[0], self.padding[0]],
                    [self.padding[1], self.padding[1]],
                    [0, 0]
                ]
            else:
                _padding = [
                    [self.padding[0], self.padding[0]],
                    [self.padding[1], self.padding[1]],
                    [0, 0]
                ]
        else:
            if len(shape) is 4:
                _padding = [
                    [self.padding[0], self.padding[0]],
                    [self.padding[1], self.padding[1]],
                    [self.padding[2], self.padding[2]],
                    [self.padding[3], self.padding[3]]
                ]
            else:
                _padding = [
                    [self.padding[0], self.padding[0]],
                    [self.padding[1], self.padding[1]],
                    [self.padding[2], self.padding[2]]
                ]

        if not self.done_first_pass:
            self.log("Padding: {}".format(_padding))
        self._data = A.pad(input, paddings=_padding)


class MergeLayer(ProcessingLayer):
    """
    Merge layers with the same shape by element-wise addition.
    """
    def _forward(self, inputs):
        shape = A.get_shape(inputs[0])
        for t in inputs:
            t_shape = A.get_shape(t)
            assert t_shape == shape,\
                "Tensors to merge do not have the same shape as {} (found {})".format(shape, t_shape)

        sum = inputs[0]
        for t in inputs[1:]:
            sum += t

        self._data = sum


class ScatterLayer(ProcessingLayer):
    """
    Scatter the input to a list of tensor according to the scattering indices
    list by the list dimension.

    Args:
        input: Tensor
            The tensor to be scattered
        scatter_list: list
            The list of length of each scattered tensor. As an example, [2,
            3, 4] will scatter input tensor to three tensors: input[0:2],
            input[2:2+3], input[5:5+4]. The last index could be safely
            omitted. If not, and the overall length has not covered the whole
            input, the remaining dimensions in the input will not be used.
    """
    def __init__(self, scatter_len_list, **kwargs):
        self.scatter_len_list = scatter_len_list
        if "name" not in kwargs:
            kwargs["name"] = "scatter"
        super(ScatterLayer, self).__init__(**kwargs)

    def _forward(self, input):
        scattered_list = []
        start_idx = 0
        for length in self.scatter_len_list:
            scattered_list.append(input[..., start_idx:start_idx+length])
            start_idx += length

        self._data = scattered_list

__all__ = [name for name, x in locals().items() if not inspect.ismodule(x)]



"""
This module holds common stuffs of all, such as global constants.
"""
from __future__ import absolute_import, print_function
import inspect
import os
import sys


# Alert if AKID_DATA_PATH is not defined.
AKID_DATA_PATH = os.getenv("AKID_DATA_PATH")
if not AKID_DATA_PATH:
    print("Environment variable AKID_DATA_PATH is not defined. It is needed to"
          " run examples.", file=sys.stderr)


# Graph collection names
# #########################################################################
DEFAULT_COLLECTION = "default"
BATCH_MONITORING_COLLECTION = "monitor"
TRAIN_SUMMARY_COLLECTION = "train_summary"
VALID_SUMMARY_COLLECTION = "val_summary"
FILTER_WEIGHT_COLLECTION = "filter_weight"
TRAINING_DYNAMICS_COLLECTION = "training_dynamics"
# Tough activation collection exists, nothing has been added to it yet. It is
# created for the sole purpose I do not want to delete the obsolete
# `get_activation` method in observer.
ACTIVATION_COLLECTION = "ACTIVATIONS"
# Auxilliary collections that may not always be used during training, such us
# eigenvalues of weight matrices (not included yet) and norm of filters are in
# this collection.
AUXILLIARY_SUMMARY_COLLECTION = "auxiliary_summary"
AUXILLIARY_STAT_COLLECTION = "auxiliary_stat"

# Global constants
# #########################################################################
# Shared seed for all involved randomness.
SEED = 66478  # Set to None for random seed.
# Manually named tag names.
LEARNING_RATE_TAG = "Learning Rate"
# tag suffixes
SPARSITY_SUMMARY_SUFFIX = "sparsity"

__all__ = [name for name, x in locals().items() if not inspect.ismodule(x)]



"""
This module contains layers that hold memories on activation. Layers in this
module are most similar to the layers in the module synapse_layers. Layers in
synapse layers learn weights, which are dual to activation, to form
hierarchical patterns, while layers in memory_layers hold memory for activation
in one layer (though it is possible the memory can be hierarchically used some
day).

The above description could be confusing, since it is not something that is
widely accepted in the deep learning community, but an unpublished, and
untested hypothesis made by me. So you may want to look at what the layers in
this module actually do, and learn how they are used in the community.
"""

from ..core.blocks import ProcessingLayer
from .. import backend as A

class MemoryLayer(ProcessingLayer):
    pass


class EmbeddingLayer(MemoryLayer):
    NAME = "Embedding"
    def __init__(self, num_embeddings, embedding_dim, *args, **kwargs):
        super(EmbeddingLayer, self).__init__(*args, **kwargs)
        self.num_embeddings = num_embeddings
        self.embedding_dim = embedding_dim

    def _setup(self):
        self.embeddings = self._get_variable("embeddings",
                                             [self.num_embeddings, self.embedding_dim],
                                             self._get_initializer())

    def _forward(self, idxs):
        self._data = A.nn.embedding(idxs, self.embeddings)
        return self._data

Embedding = EmbeddingLayer



from __future__ import absolute_import
from akid import AKID_DATA_PATH
from akid.core import kids, kongfus
from akid import Cifar10TFSource
from akid import IntegratedSensor
from akid import WhitenJoker

from akid.models.brains import VGGNet

# Set up brain
# #########################################################################
brain = VGGNet(padding="SAME",
               name='torch-vgg-cifar10')

# Set up a sensor.
# #########################################################################
cifar_source = Cifar10TFSource(
    name="CIFAR10",
    url='http://www.cs.toronto.edu/~kriz/cifar-10-binary.tar.gz',
    work_dir=AKID_DATA_PATH + '/cifar10',
    num_train=50000,
    num_val=10000)


sensor = IntegratedSensor(source_in=cifar_source,
                          batch_size=128,
                          name='data')
sensor.attach(WhitenJoker(name="per_image_whitening"))
sensor.attach(WhitenJoker(name="per_image_whitening"), to_val=True)

# Summon a survivor.
# #########################################################################
survivor = kids.Kid(
    sensor,
    brain,
    kongfus.MomentumKongFu(base_lr=1,
                           decay_rate=0.1,
                           decay_epoch_num=25),
    max_steps=510000)
survivor.setup()

# Start training
# #######################################################################
survivor.practice()



"""Operations often used for initializing tensors."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import inspect

from tensorflow.python.framework import dtypes
from tensorflow.python.ops import random_ops


def msra_initializer(factor=1.0, uniform=False, seed=None):
    """Returns an initializer that implement MSRA initialization. Refer to the
    [paper](http://arxiv.org/abs/1502.01852), *Delving Deep into Rectifiers:
    Surpassing Human-Level Performance on ImageNet Classification*, for
    details.

    It uses the version that keeps gradient variance 1. For the version to
    forward propagation variance 1, the `uniform_unit_scaling_initializer`
    provided by tensorflow suffices --- what you need to do is to change the
    multiplicative factor.

    Also, since the constant in MSRA init is also approximated in its
    derivation, a multiplication factor parameter is also introduced.

    Args:
        factor: Float.  A multiplicative factor by which the values will be
            scaled.
        seed: A Python integer. Used to create random seeds.

    Returns:
        An initializer that generates tensors with unit variance.
    """
    def _initializer(shape, dtype=dtypes.float32, partition_info=None):
        input_size = 1.0
        for i in shape[:-2]:
            input_size *= i
        input_size *= shape[-1]
        std = math.sqrt(2 / input_size) * factor
        if uniform:
            max_val = std
            return random_ops.random_uniform(shape, -max_val, max_val,
                                             dtype, seed=seed)
        else:
            return random_ops.truncated_normal(shape, stddev=std,
                                               dtype=dtype, seed=seed)
    return _initializer


__all__ = [name for name, x in locals().items() if not inspect.ismodule(x)]



"""
This module contains different genres of Kong Fu to a `Kid` to practice,
aka different training algorithms and policies to train a network.
"""
from __future__ import absolute_import
import abc
import inspect

import tensorflow as tf

from .common import TRAINING_DYNAMICS_COLLECTION, LEARNING_RATE_TAG
from .blocks import ShadowableBlock
from .interface_blocks import UpdateBlock
from .. import backend as A


class LearningRateScheme(object):
    exp_decay = 1
    placeholder = 2

class KongFu(ShadowableBlock, UpdateBlock):
    """
    An top level abstract class to compute gradients given a loss.  Any
    concrete `KongFu` should implement `_get_optimizer` to provide a concrete
    optimizer.

    To change learning rate call `set_lr`. For example, to set a learning rate
    scheduler during training::

        kid = Kid(...)
        from akid import backend as A

        def update_lr(kid):
            if A.get_step() < 200:
                kid.kongfu.set_lr(0.1)
            elif A.get_step() < 400:
                kid.kongfu.set_lr(0.01)
            elif A.get_step() < 600:
                kid.kongfu.set_lr(0.001)
            else:
                kid.kongfu.set_lr(0.0001)

        kid.hooks.on_batch_begin.append(update_lr)

    The data of this block depends on backends.

    For PyTorch, since the gradients are saved along with variable
    tensors. `data` property is None.

    For Tensorflow: all concrete sub-class should save the computed gradients
    to `_data`, which is exposed by the property `data`. The data should be the
    return of `tf.train.Optimizer.compute_gradients`, which is a list of
    (gradient, variable) pairs. Further processing of the gradients could be
    done anyway you want, such as doing gradient average for data parallelism,
    or gradient clipping.
    """
    NAME = "Kongfu"

    def __init__(self,
                 lr=0.01,
                 var_list=None,
                 lr_scheme={"name": LearningRateScheme.exp_decay,
                            "base_lr": 0.01,
                            "decay_rate": 0.95,
                            "num_batches_per_epoch": 468,
                            "decay_epoch_num": 1},
                 **kwargs):
        """
        Args:
            var_list: list
                The list of variables that are supposed to train.
            lr: float
                The learning rate.
            lr_scheme: dict.
                This option is deprecated. Use lr instead. This option is
                supported with 'tensorflow' backend. To use it, `lr` needs to
                be set to None explicitly.

                Learning rate scheme to use. Two types are supported for now:
                    1. `exp_decay`: exponential decay. 'base_lr',
                       'decay_rate' are required parameters. Either
                       'decay_steps' or 'decay_epoch_num' should be
                       present. If using 'decay_epoch_num', an additional
                       parameter 'num_batches_per_epoch' should be passed
                       in. It is used to convert epoch number to step number.
                    2. `placeholder`: a placeholder that is supposed to pass
                       in by feed dict, so arbitrary hard coded learning rate
                       decay could be used. To use this scheme, you should
                       update `lr_value` of `KongFu` manually, which is
                       normally done by assigning value in some callback
                       functions. Alternatively, if you are using `KongFu`
                       standalone, you need to feed a value to
                       `KongFu._lr_tensor`.
                See the default value for an example usage.
        """
        super(KongFu, self).__init__(**kwargs)
        self._lr_value = lr
        if var_list is None:
            self.var_list = []
        else:
            self.var_list = var_list

        # Deprecated. Save for compatibility.
        if lr is None:
            self.lr_scheme = lr_scheme
        else:
            self.lr_scheme = None

    def _setup(self):
        if self.lr_scheme:
            # For compatibility.
            self._setup_lr_scheme()
        else:
            if A.backend() == A.TF:
                self._lr_tensor = tf.placeholder(tf.float32,
                                            shape=[],
                                            name='lrn')
                self.opt = self._get_optimizer(self._lr_tensor)
            elif A.backend() == A.TORCH:
                self.opt = self._get_optimizer(self._lr_value)
                A.cache_tensor(self._lr_value, LEARNING_RATE_TAG)

    def set_lr(self, lr):
        """
        Set the learning rate.
        """
        self._lr_value = lr

        if A.backend() == A.TORCH:
            if self.is_setup:
                for pg in self.opt.param_groups:
                    pg['lr'] = lr
            A.cache_tensor(self._lr_value, LEARNING_RATE_TAG)

    def set_var_list(self, var_list):
        self.var_list = var_list

    def append_var_list(self, var_list):
        self.var_list.extend(var_list)

    def get_lr(self):
        return self._lr_value

    def _setup_lr_scheme(self):
        if A.backend() != A.TF:
            raise ValueError("'lr_scheme' is supported only with tensorflow backend.")

        if self.lr_scheme["name"] is LearningRateScheme.exp_decay:
            base_lr = float(self.lr_scheme["base_lr"])
            decay_rate = self.lr_scheme["decay_rate"]
            if "decay_steps" not in self.lr_scheme:
                decay_epoch_num = self.lr_scheme["decay_epoch_num"]
                num_batches_per_epoch \
                    = self.lr_scheme["num_batches_per_epoch"]
                decay_steps = num_batches_per_epoch * decay_epoch_num
            else:
                decay_steps = self.lr_scheme["decay_steps"]

            # TODO: step is not a tensor now, it may not be incremented during
            # optimization, which means the exponential decay will be broken ...
            step = A.get_step()

            self._lr_tensor = tf.train.exponential_decay(
                base_lr,
                step,
                decay_steps,
                decay_rate,
                staircase=True)
        elif self.lr_scheme["name"] is LearningRateScheme.placeholder:
            self._lr_tensor = tf.placeholder(tf.float32,
                                        shape=[],
                                        name='lrn')
            self._lr_value = None
        else:
            raise Exception("Learning rate scheme is not supported. Please"
                            " specify one from `LearningRateScheme`.")

    def _forward(self, loss):
        """
        Build and return training ops according to the loss.
        """
        self._data = A.train.compute_gradients(self.opt, loss)
        return self._data

    def _update(self, grads):
        return A.train.apply_gradients(self.opt, grads)

    def _post_forward(self, *args, **kwargs):
        if self.done_first_pass:
            return

        if self.do_summary:
            if A.backend() == A.TF:
                A.summary.scalar(LEARNING_RATE_TAG,
                                self._lr_tensor,
                                collections=[TRAINING_DYNAMICS_COLLECTION])
            elif A.backend() == A.TORCH:
                A.summary.scalar(LEARNING_RATE_TAG,
                                self._lr_value,
                                collections=[TRAINING_DYNAMICS_COLLECTION])

    def get_feed_dict(self):
        """
        Return a dict that fills the learning rate placeholder with current
        learning rate. Only useful when the backend is tensorflow.
        """
        return {self._lr_tensor: self._lr_value}

    @property
    def data(self):
        return self._data

    @abc.abstractmethod
    def _get_optimizer(self, lr):
        """
        An abstract method to get an optimizer. Sub-class of `KongFu` should
        implement this method to return a concrete optimizer for training.

        Args:
            lr: tensor variable returned by tf.train.exponential_decay
                The periodically decayed learning rate.
        """
        raise NotImplementedError('Each sub-kongfu needs to implement this'
                                  'method to provide an optimizer!')


class MomentumKongFu(KongFu):
    def __init__(self, momentum=0.9, use_nesterov=False, **kwargs):
        super(MomentumKongFu, self).__init__(**kwargs)
        self.momentum = momentum
        self.use_nesterov = use_nesterov

    def _get_optimizer(self, lr):
        return A.train.MomentumOptimizer(lr,
                                         self.var_list,
                                         momentum=self.momentum,
                                         use_nesterov=self.use_nesterov)


class GradientDescentKongFu(KongFu):
    def _get_optimizer(self, lr):
        return A.train.GradientDescentOptimizer(lr,
                                                self.var_list)


class AdamKongFu(KongFu):
    def _get_optimizer(self, lr):
        return A.train.AdamOptimizer(lr)


__all__ = [name for name, x in locals().items() if
           not inspect.ismodule(x) and not inspect.isabstract(x)]



from __future__ import absolute_import
import tensorflow as tf
from . import computational_graph as cg
from akid.core.common import SEED
from akid.utils import glog as log
from six.moves import range


def depthwise_conv2d(input, filter, bias=None, strides=1, padding='VALID', name=None):
    strides = expand_kernel(strides)
    v = tf.nn.depthwise_conv2d(input, filter, strides, padding, name)

    if bias:
        v = tf.nn.bias_add(v, bias)

    return v


def max_pool(value, ksize, strides, padding, data_format="NHWC", name=None):
    ksize = expand_kernel(ksize)
    strides = expand_kernel(strides)
    return tf.nn.max_pool(value, ksize, strides, padding, data_format, name)


def avg_pool(value, ksize, strides, padding, data_format="NHWC", name=None):
    ksize = expand_kernel(ksize)
    strides = expand_kernel(strides)
    return tf.nn.avg_pool(value, ksize, strides, padding, data_format, name)


def max_pool_with_argmax(input, ksize, strides, padding, Targmax=None, name=None):
    ksize = expand_kernel(ksize)
    strides = expand_kernel(strides)
    return tf.nn.max_pool_with_argmax(input, ksize, strides, padding, Targmax=Targmax, name=name)


def _max_unpooling(X_in, mask, ksize=2, name="unpool"):
    """
    Args:
        X_in: Tensor
            Input to unpool.
        mask:
            Indices output by `tf.nn.max_pool_with_argmax.

    Adopted from: https://github.com/tensorflow/tensorflow/issues/2169
    """
    if isinstance(ksize, int):
        ksize = [1, ksize, ksize, 1]
    input_shape = X_in.get_shape().as_list()
    #  calculation new shape
    output_shape = [input_shape[1] * ksize[1], input_shape[2] * ksize[2], input_shape[3]]
    # calculation indices for batch, height, width and feature maps
    one_like_mask = tf.ones_like(mask)
    bsize = tf.to_int64(tf.shape(X_in)[0])
    batch_range = tf.reshape(tf.range(bsize, dtype=tf.int64),
                             shape=[-1, 1, 1, 1])
    b = one_like_mask * batch_range
    y = mask // (output_shape[1] * output_shape[2])
    x = mask % (output_shape[1] * output_shape[2]) // output_shape[2]
    feature_range = tf.range(output_shape[2], dtype=tf.int64)
    f = one_like_mask * feature_range
    # transpose indices & reshape update values to one dimension
    updates_size = tf.size(X_in)
    indices = tf.transpose(tf.reshape(tf.stack([b, y, x, f]), [4, updates_size]))
    values = tf.reshape(X_in, [updates_size])
    ret = tf.scatter_nd(indices, values, tf.concat(
        [[bsize], tf.to_int64(output_shape)], axis=0))
    return ret


def max_unpooling(X_in, mask, ksize=[1, 2, 2, 1]):
    """
    Does not support partially defined shape. For a tentative version that
    supports dynamic batch size, refer to `_max_pooling`.
    """
    ksize = expand_kernel(ksize)
    input_shape = X_in.get_shape().as_list()
    #  calculation new shape
    output_shape = (input_shape[0], input_shape[1] * ksize[1], input_shape[2] * ksize[2], input_shape[3])
    # calculation indices for batch, height, width and feature maps
    one_like_mask = tf.ones_like(mask)
    batch_range = tf.reshape(tf.range(output_shape[0], dtype=tf.int64), shape=[input_shape[0], 1, 1, 1])
    b = one_like_mask * batch_range
    y = mask // (output_shape[2] * output_shape[3])
    x = mask % (output_shape[2] * output_shape[3]) // output_shape[3]
    feature_range = tf.range(output_shape[3], dtype=tf.int64)
    f = one_like_mask * feature_range
    # transpose indices & reshape update values to one dimension
    updates_size = tf.size(X_in)
    indices = tf.transpose(tf.reshape(tf.stack([b, y, x, f]), [4, updates_size]))
    values = tf.reshape(X_in, [updates_size])
    X_out = tf.scatter_nd(indices, values, output_shape)

    return X_out


def relu(X_in, name=None):
    return tf.nn.relu(X_in, name)


def conv2d(input, filter, bias=None, strides=1, padding="VALID", name=None):
    strides = expand_kernel(strides)

    if bias is not None:
        v = tf.nn.conv2d(input, filter,
                         strides, padding,
                         use_cudnn_on_gpu=None, data_format=None)
        v = tf.nn.bias_add(v, bias, name=name)
    else:
        v = tf.nn.conv2d(input, filter,
                         strides, padding,
                         use_cudnn_on_gpu=None, data_format=None,
                         name=name)

    return v


def inner_product(input, W, bias=None, name=None):
    if bias is not None:
        ip = tf.matmul(input, W)
        ip = tf.nn.bias_add(ip, bias, name=name)
    else:
        ip = tf.matmul(input, W, name=name)

    return ip


def l2_loss(var):
    return tf.nn.l2_loss(var)


def l1_loss(var):
    return tf.reduce_sum(tf.abs(var))


def zero_fraction(data, name=None):
    return tf.nn.zero_fraction(data, name=name)


def mse_loss(data, labels, size_average=None, name=None):
    """
    size_average is of no use, to be compatible with PyTorch backend.
    """
    if size_average == False:
        reduction_type = tf.losses.Reduction.NONE
    else:
        reduction_type = tf.losses.Reduction.MEAN

    return tf.losses.mean_squared_error(labels, data, scope=name, reduction=reduction_type)


def cross_entropy_loss(logits, labels, name=None):
    label_shape_size = len(labels.get_shape().as_list())
    class_num = cg.get_shape(logits)[1]
    if label_shape_size is 1:
        # Set up loss graph.
        # Convert from sparse integer labels in the range [0, NUM_CLASSSES)
        # to 1-hot dense float vectors (that is we will have batch_size
        # vectors, each with NUM_CLASSES values, all of which are 0.0
        # except there will be a 1.0 in the entry corresponding to the
        # label).
        batch_size = tf.size(labels)
        _labels = tf.expand_dims(labels, 1)
        indices = tf.expand_dims(tf.range(0, batch_size, 1), 1)
        concated = tf.concat(axis=1, values=[indices, _labels])
        onehot_labels = tf.sparse_to_dense(
            concated, tf.stack([batch_size, class_num]), 1.0, 0.0)
    else:
        # The labels has already been expanded, no need to do it again.
        onehot_labels = labels

    cross_entropy \
        = tf.nn.softmax_cross_entropy_with_logits(logits=logits,
                                                    labels=onehot_labels,
                                                    name='xentropy')
    cross_entropy_mean = tf.reduce_mean(cross_entropy,
                                        name=name)

    return cross_entropy_mean


def class_acccuracy(predictions, labels, name=None):
    label_shape_size = len(cg.get_shape(labels))
    if label_shape_size is 1:
        # NOTE: this approach has a bug. If the classifier is so weak, it
        # gives all equal probability for all classes, then the final
        # accuracy would be 1, since obviously the desired class is in the
        # top k.
        correct = tf.nn.in_top_k(predictions, labels, 1)
        # Return the number of true entries.
    else:
        truth = tf.argmax(labels, axis=1)
        predictions = tf.argmax(predictions, axis=1)
        correct = tf.equal(truth, predictions)

    return tf.reduce_mean(tf.cast(correct, tf.float32),
                          name=name)


def dropout(v, keep_prob, val=False, in_place=False, name=None):
    if in_place:
        raise ValueError("In place is not supported in tensorflow backend.")

    if val:
        return tf.identity(v, name=name)
    else:
        return tf.nn.dropout(v, keep_prob, seed=SEED, name=name)


def expand_kernel(ksize):
    if type(ksize) == int:
        ksize = [1, ksize, ksize, 1]
    elif len(ksize) == 2:
        ksize = [1, ksize[0], ksize[1], 1]
    else:
        ksize = ksize

    return ksize


def bn(x, out_channel_num, gamma, beta,
       is_val=False, step=None, fix_gamma=False, share_gamma=False,
       name=None):
    # TODO: moving average track is not optional as in the torch backend.
    def _bn(input, mean, variance, beta, gamma, epsilon):
        shape = input.get_shape().as_list()
        if len(shape) is 2 or share_gamma:
            normalized_input = (input - mean) / tf.sqrt(variance + epsilon)
            normalized_input *= gamma
            out = tf.add(normalized_input,
                         beta,
                         name=name)
        else:
            out = tf.nn.batch_norm_with_global_normalization(
                input, mean, variance, beta,
                gamma, epsilon, fix_gamma, name=name)

        return out

    input_shape = cg.get_shape(x)
    reduction_axes = list(range(len(input_shape)))
    del reduction_axes[-1]
    mean, variance = tf.nn.moments(x, reduction_axes)

    # Bookkeeping a moving average for inference.

    # Since the initial mean and average are not accurate, we should use a
    # lower lower momentum. This is particularly important for ResNet since
    # the initial activation could be very large due to the exponential
    # accumulation effect of merge layers, though it does not work not well
    # to remove the effect for ResNet. To achieve this, we use the
    # mechanism provided by tensorflow, by passing current step in.
    ema = tf.train.ExponentialMovingAverage(0.9, step)
    with tf.variable_scope(tf.get_variable_scope(), reuse=False):
        # NOTE: Prior to tf 0.12, I did not need the variable scope above
        # this line to get things working. The problem is that moving
        # average cannot be put in a variable scope that plans to reuse its
        # variables (due to a debias mechanism introduced in tf 0.12). This
        # is a temporary fix, waiting for solutions in issues:
        # https://github.com/tensorflow/tensorflow/issues/6270 and
        # https://github.com/tensorflow/tensorflow/issues/5827
        ema_apply_op = ema.apply([mean, variance])
    ema_mean, ema_var = ema.average(mean), ema.average(variance)
    with tf.control_dependencies(
            [ema_apply_op]):
        if is_val:
            out = _bn(x, ema_mean, ema_var, beta, gamma, 1e-5)
        else:
            out = _bn(x, mean, variance, beta, gamma, 1e-5)

    return out



class UpdateBlock(object):
    """
    A block that supports an `update` method to update any parameters in the
    block.
    """
    def __init__(self, call_on_update=False, **kwargs):
        super(UpdateBlock, self).__init__(**kwargs)
        self.call_on_update = call_on_update

    def update(self, *args, **kwargs):
        d = self._update(*args, **kwargs)

        if self.call_on_update:
            self._on_update(*args, **kwargs)

        return d

    def _update(self):
        raise NotImplementedError("The update method is not implemented."
                                  "The program should not reach here.")

    def _on_update(self, *args, **kwargs):
        """
        Hook to be called after parameter update.
        """
        pass



"""
Event to pass around to communicate events happening in blocks.
"""
class DataPrefetchThreadsDeadEvent(Exception):
    pass


class DataPrefetchProcessesDeadEvent(Exception):
    pass


class EarlyStoppingEvent(Exception):
    def __init__(self, val_loss, val_evals):
        """
        When the training early stops, it should save current validation loss
        and evaluations.
        """
        self.val_loss = val_loss
        self.val_evals = val_evals


class EpochCompletedEvent(Exception):
    pass


class DoneEvent(Exception):
    pass



from __future__ import print_function

from __future__ import absolute_import
import os
import sys
import inspect

from akid.ops import image_ops as image

from akid.core.sources import *
from akid import datasets
from akid.datasets import *
from akid.core.kids import *
from akid.core.kongfus import *
from akid.core.sensors import *
from akid.core.jokers import *
from akid.core.common import *
from akid.core.brains import *
from akid.core.eval_blocks import *
from akid.core import events
from akid.core import initializers
from akid.core import common
from akid.core import engines
from akid.core import samplers
try:
    from akid.core.observer import *
except ImportError as e:
    print("Cannot import observer. You probably run on a machine without"
          " matplotlib.")

# Expose sub-modules
import akid.nn as nn
import akid.layers as layers
from akid import backend as A


def reset():
    A.reset()
    reset_eval_block_map()


__all__ = [name for name, x in locals().items() if
           not inspect.ismodule(x) and not inspect.isabstract(x)]



#!/usr/bin/env python

from __future__ import print_function

from __future__ import absolute_import
import argparse
import operator

import tensorflow as tf

from akid import LEARNING_RATE_TAG

parser = argparse.ArgumentParser(
    description="""
Description:
# #########################################################################
Given a tensorflow event file, print out the top 10 accuracy during training.
    """,
    formatter_class=argparse.RawDescriptionHelpFormatter)
parser.add_argument("EVENT_FILE",
                    type=str,
                    help="path to the event file.")
parser.add_argument("eval_metric",
                    type=str,
                    help="""
                    The evaluation metric to evaluate the model. Since a model
                    may have multiple evaluation metrics associated with it,
                    and normally only one of them is critical (at least for
                    now), it is supposed to be told. The name is the name of an
                    evaluation metric in the training log. A full name is not
                    needed as long as the part provided is unique enough to
                    distinguish it from other metrics (otherwise, the last
                    match will be used, which normally will not be the
                    desirable behavior).
                    """)
parser.add_argument("-s", "--step",
                    type=int,
                    help="""
                    The start step to check. The reason that this argument
                    exists is, in certain situation, accuracy reported is wrong
                    at the beginning of the training, thus should not be
                    counted when trying to find the max ones.
                    """)
arguments = parser.parse_args()

EVENT_FILE_PATH = arguments.EVENT_FILE
eval_metric_name = arguments.eval_metric
start_step = arguments.step

# Gather the list of validation accuracy.
val_acc_tuples = []
previous_step = None
for e in tf.train.summary_iterator(EVENT_FILE_PATH):
    if e.step < start_step:
        continue
    # Since a step may be logged by multiple event files, state variables are
    # only updated when we are switching steps.
    if e.step != previous_step:
        val_acc = None
        lr = None
        previous_step = e.step

    for v in e.summary.value:
        if v.tag.find(eval_metric_name) is not -1:
            val_acc = v.simple_value
        if v.tag == LEARNING_RATE_TAG:
            lr = v.simple_value
    if val_acc and lr:
        val_acc_tuples.append((e.step, val_acc, lr))


# Find and print the top 10.
sorted_val_acc_tuples = sorted(val_acc_tuples,
                               key=operator.itemgetter(1),
                               reverse=True)
for i, t in enumerate(sorted_val_acc_tuples[0:10]):
    print("Rank {} : iter {}, accuracy ({}) {:.5f}, lr {:.8f}".format(
        i, t[0], eval_metric_name, t[1], t[2]))



from __future__ import absolute_import
def setup():
    from akid import AKID_DATA_PATH
    from akid import Kid
    from akid import LearningRateScheme, GradientDescentKongFu
    from akid import Cifar10TFSource
    from akid import IntegratedSensor
    from akid import (
        CropJoker,
        WhitenJoker,
        FlipJoker,
        LightJoker
    )

    from akid.models.brains import AlexNet

    # Set up brain
    # #########################################################################
    brain = AlexNet(in_channel_num=2304, dataset="cifar10", name='alex_net')

    # Set up a sensor.
    # #########################################################################
    cifar_source = Cifar10TFSource(
        name="CIFAR10",
        url='http://www.cs.toronto.edu/~kriz/cifar-10-binary.tar.gz',
        work_dir=AKID_DATA_PATH + '/cifar10',
        num_train=50000,
        num_val=10000)


    sensor = IntegratedSensor(source_in=cifar_source,
                              batch_size=128,
                              name='data')
    sensor.attach(CropJoker(height=24, width=24, center=True, name="crop"),
                  to_val=True)
    sensor.attach(WhitenJoker(name="per_image_whitening"), to_val=True)

    sensor.attach(CropJoker(height=24, width=24, center=False, name="crop"))
    sensor.attach(FlipJoker(name="left_right_flip"))
    sensor.attach(LightJoker(name="brightness_contrast"))
    sensor.attach(WhitenJoker(name="per_image_whitening"))

    # Summon a survivor.
    # #########################################################################
    kid = Kid(
        sensor,
        brain,
        GradientDescentKongFu(
            lr_scheme={
                "name": LearningRateScheme.exp_decay,
                "base_lr": 0.1,
                "decay_rate": 0.1,
                "num_batches_per_epoch": 391,
                "decay_epoch_num": 350}),
        max_steps=510000)
    kid.setup()

    return kid

# Start training
# #######################################################################
if __name__ == "__main__":
    kid = setup()
    kid.practice(continue_from_chk_point=False)



"""
Tensorflow backend for akid.

NOTE
# #########################################################################
To use tensorflow backend to execute/evaluate any graph built, run `init`
beforehand.
"""
from __future__ import absolute_import
import sys

from uuid import uuid4
import numpy as np
import tensorflow as tf

from .. import computational_graph as cg
from akid.utils import glog as log


DATA_FORMAT = "HWC"
# Build a global session to use
# TODO: think how to offer a way to close the session.
# A bug in tensorflow: Exception AttributeError: "'NoneType' object has no attribute 'TF_DeleteStatus'" in <bound method Session.__del__ of <tensorflow.python.client.session.Session object at 0x7f43105e9dd0>> ignored
# Should be fixed by now: https://github.com/tensorflow/tensorflow/issues/3388
sess = None
saver = None
initialized = False

float32 = tf.float32


def init(continue_from_chk_point=False, model_dir=None):
    global sess
    global saver
    global initialized

    if not initialized:
        if len(tf.global_variables()) != 0:
            saver = tf.train.Saver(tf.global_variables())

        config = tf.ConfigProto(allow_soft_placement=True)
        config.gpu_options.allow_growth = True
        sess = tf.Session(config=config)

        if continue_from_chk_point:
            log.info("Recovering net from checkpoint %s." % model_dir)
            restore(model_dir)
        else:
            init  = tf.group(tf.global_variables_initializer(),
                            tf.local_variables_initializer())
            sess.run(init)

        tf.train.start_queue_runners(sess=sess)

        initialized = True


def close():
    if sess is not None:
        sess.close()

    global initialized
    initialized = False


def get_variable(name=None, shape=None,
                 initializer=None, trainable=True,
                 shared=True):
    """
    Refer to the same function in PyTorch backend for semantics.
    """
    # Generate a random name if the name is not given
    if not name:
        name = get_random_name()

    if not callable(initializer):
        shape = None

    if shared:
        if hasattr(initializer, "native"):
            # Use naive initializer instead of that of tensorflow
            init = initializer(shape)
            shape = None
        else:
            init = initializer

        return tf.get_variable(name, shape,
                               initializer=init, trainable=trainable)
    else:
        raise NotImplementedError("Normal Variable creation has not been implemented yet.")


def get_all_variables():
    return tf.global_variables()


def save(path):
    saver.save(sess,
               path + "/checkpoint",
               global_step=cg.get_step())


def restore(path):
    checkpoint = tf.train.get_checkpoint_state(path)
    if checkpoint and checkpoint.model_checkpoint_path:
        saver.restore(sess, checkpoint.model_checkpoint_path)
        filename = checkpoint.model_checkpoint_path.split('/')[-1]
        step = int(filename.split('-')[-1])
        cg.set_step(step)
    else:
        log.error("No checkpoint found under %s!" % path)
        sys.exit()



def get_name(v):
    """
    Refer to the same function of PyTorch backend for docs.
    """
    if v is not None:
        return v.op.name.split(':')[0]
    else:
        return None


def remove_scope_from_name(name):
    """
    Remove scope and device information in the name.
    """
    name = name.split('/')[-1]

    return name


def append_suffix(name, suffix):
    """
    Append suffix to the name of a tensor, above the level of devices.
    """
    return name + '/' + suffix


def Tensor(X_in, requires_grad=False):
    """
    Get a tensor from a constant (array).

    Args:
        requires_grad: It is for compatibility with PyTorch. Not used here,
            since in tensorflow tensor can participate in auto-differentiation.
    """
    return tf.constant(X_in, dtype=tf.float32)


def eval(V):
    """
    Convert variable to numpy array.
    """
    if type(V) is np.ndarray:
        return V

    return sess.run(V)


def run(op, feed_dict=None):
    """Run an operation with arguments. For compatibility with Tensorflow"""
    return sess.run(op, feed_dict=feed_dict)


def split(split_dim, num_split, value, name="split"):
    return tf.split(axis=split_dim, num_or_size_splits=num_split, value=value, name="split")


def reshape(tensor, shape, name=None):
    return tf.reshape(tensor, shape, name)


def reduce_sum(input_tensor, axis=None, keep_dims=False, name=None, reduction_indices=None):
    tf.reduce_sum(input_tensor, axis, keep_dims, name, reduction_indices)


def cast(x, dtype, name=None):
    return tf.cast(x, dtype, name)


def concat(concat_dim, values, name="concat"):
    return tf.concat(axis=concat_dim, values=values, name=name)


def stack(values, axis=0, name="pack"):
    return tf.stack(values, axis=axis, name=name)


def unstack(values, axis=0, name="unstack"):
    return tf.unstack(values, num=None, axis=axis, name=name)


def pack(values, axis=0, name="pack"):
    return tf.stack(values, axis=axis, name=name)


def expand_dims(input, axis=None, name=None):
    return tf.expand_dims(input, axis=axis, name=name)


def reduce_max(input_tensor, axis=None, keep_dims=False, name=None):
    return tf.reduce_max(input_tensor, axis=axis, keep_dims=keep_dims, name=name)


def abs(x, name=None):
    return tf.abs(x, name=name)


def is_tensor(T):
    return type(T) is tf.Tensor


def mul(a, b, name=None):
    return tf.multiply(a, b, name)


def mean(v, name=None):
    return tf.reduce_mean(v, name=name)


def add_n(l, name=None):
    return tf.add_n(l, name=name)


def div(v, denominator, name=None):
    return tf.div(v, denominator, name=name)


def zero_fraction(data):
    return tf.nn.zero_fraction(data)


def get_shape(t):
    return t.get_shape().as_list()


def standardize_data_format(data, old_format):
    """
    Stardardize data to Tensorflow format, which is first last.

    Args:
        data: Tensor or numpy array.
            The input data.
        old_format: str
            A string describe the original format. For example, if converting
            from Tensorflow, it would be 'hwio' for parameter. See
            `SUPPORT_DATA_FORMAT` and `SUPPORT_PARA_FORMAT` for supported
            strings.
    """
    if old_format not in cg.SUPPORT_PARA_FORMAT \
       and old_format not in cg.SUPPORT_DATA_FORMAT:
        raise ValueError("The data format {} is not well specified.".format(old_format))

    if old_format in cg.SUPPORT_PARA_FORMAT:
        out_format = 'hwio'
    else:
        out_format = 'nhwc'

    if type(data) == np.ndarray:
        return np.einsum('{}->{}'.format(old_format, out_format), data)
    else:
        raise ValueError("Type {} is not supported.".format(type(data)))


def get_random_name():
    return uuid4().hex


def scatter(data, devices, name=None):
    # Since tensorflow handles all communications, just split the data with the
    # right number.
    num_split = len(devices)
    return tf.split(axis=0, num_or_size_splits=num_split, value=data, name=name)



"""
A collection of functions that may attach to hooks of `Kid`.

Each function is expected to take a `Kid` instance as an input that is
supposed to hold all information needed.
"""
from __future__ import absolute_import
from akid import backend as A
import akid as K
from akid import ops

from ..utils import glog as log
from ..utils.tools import is_tuple_or_list
from six.moves import zip


def _do_summary(kid):
    if kid.do_summary:
        if A.backend() == A.TF:
            feed_dict = kid.feed_dict
        elif A.backend() == A.TORCH:
            if kid.do_summary_on_val:
                kid.sensor.set_mode("val")
                kid.sensor.setup()
                # If doing summary on validation set, a step needs to be run to
                # update the feature maps. Tensorflow will do this
                # automatically, so forward once only for torch.
                kid.step(update=False, val=True)
                kid.sensor.set_mode("train")
                kid.sensor.setup()

            feed_dict = None

        A.summary.run_summary_op(kid.summary_op, feed_dict=feed_dict)


def v_to_text(name_list, v_list):
    """
    Convert named tensors to text to print. v is a list of tensors.
    """
    return [K.get_eval_block(name_list[i]).__str__(None, v) for i, v in enumerate(v_list)]


def on_train_log_step(kid):
    loss = kid.loss
    evals = kid.evals
    duration = kid.step_time
    step = A.get_step()

    kid.loss_data_train.append(loss)
    kid.eval_data_train.append(evals)

    name_to_print = [A.get_name(g, no_scope=True) for g in kid.engine.eval()]
    eval_value_to_print = v_to_text(name_to_print, evals)
    eval_to_print = dict(list(zip(name_to_print, eval_value_to_print)))

    num_examples_per_step = kid.sensor.batch_size
    examples_per_sec = num_examples_per_step / duration
    sec_per_batch = float(duration)

    lr = kid.kongfu.get_lr()

    log.info("Step {}: loss = {:.5f} lr = {:.8f} acc = {} ({:.1f}"
             " examples/sec {:.3f} sec/batch)".format(
                 step,
                 loss,
                 lr,
                 eval_to_print,
                 examples_per_sec,
                 sec_per_batch))

    _do_summary(kid)

    if kid.do_batch_monitoring:
        kid.batch_monitoring(kid.cached_data)


def on_val_log_step(kid):
    if kid.do_summary:
        # Add summary.
        A.summary.add_scalar(name="Validation Loss",
                             value=kid.loss,
                             step=A.get_step())
        for i, v in enumerate(kid.evals):
            if A.is_numerical(v):
                A.summary.add_scalar(
                    name=A.append_suffix(A.get_name(kid.engine.eval(get_val=True)[i]), "val"),
                    value=v,
                    step=A.get_step())
            elif is_tuple_or_list(v):
                for j, u in enumerate(v):
                    A.summary.add_scalar(
                        name=A.append_suffix(A.get_name(kid.engine.eval(get_val=True)[i][j]), "val_{}".format(i)),
                        value=u,
                        step=A.get_step())
        if kid.verbose_evals is not None:
            for i, v in enumerate(kid.verbose_evals):
                if A.is_numerical(v):
                    A.summary.add_scalar(
                        name=A.append_suffix(A.get_name(kid.engine.verbose_eval(get_val=True)[i]), "vval"),
                        value=v,
                        step=A.get_step())
                elif is_tuple_or_list(v):
                    for j, u in enumerate(v):
                        A.summary.add_scalar(
                            name=A.append_suffix(A.get_name(kid.engine.verbose_eval(get_val=True)[i]), "vval_{}".format(j)),
                            value=u,
                            step=A.get_step())

    # Log current validation.
    name_to_print = [A.get_name(g, no_scope=True) for g in kid.engine.eval(get_val=True)]
    eval_value_to_print = v_to_text(name_to_print, kid.evals)
    eval_to_print = dict(list(zip(name_to_print, eval_value_to_print)))
    if kid.verbose_evals is not None:
        name_to_print = [A.get_name(g, no_scope=True) for g in kid.engine.verbose_eval(get_val=True)]
        veval_value_to_print =  v_to_text(name_to_print, kid.verbose_evals)
        veval_to_print = dict(list(zip(name_to_print, veval_value_to_print)))
    else:
        veval_to_print = None
    log.info('  Num examples: {}  Evals : {}  VEvals: {}'.format(
        kid.sensor.source.size, eval_to_print, veval_to_print))

    # Log current best validation.
    name_to_print = [A.get_name(g, no_scope=True) + '_best'
                     for g in kid.engine.eval(get_val=True)]
    eval_value_to_print = v_to_text([A.get_name(g, no_scope=True) for g in kid.engine.eval(get_val=True)],
                                    kid.best_val_evals)
    eval_to_print = dict(list(zip(name_to_print, eval_value_to_print)))
    log.info('Current best evals : {}'.format(eval_to_print))

    # Loss.
    log.info('  Step %d: Validation loss = %.6f' % (A.get_step(), kid.loss))

    kid.loss_data_val.append(kid.loss)
    kid.eval_data_val.append(kid.evals)


def on_train_begin(kid):
    # Calculate and log total parameter number.
    total_parameters = kid.brain.num_parameters()
    log.info("Total parameters: {}".format(total_parameters))

    # Set up summary ops
    kid.setup_summary()

    # Run ops once to show initial training loss and save initial
    # summaries.
    kid.loss, kid.evals = kid.run_step(update=False)

    _do_summary(kid)

    name_to_print = [A.get_name(g, no_scope=True) for g in kid.engine.eval()]
    eval_value_to_print = v_to_text(name_to_print, kid.evals)
    eval_to_print = dict(list(zip(name_to_print, eval_value_to_print)))
    log.info("Step {}: loss = {:.5f} eval = {}".format(
        A.get_step(), kid.loss, eval_to_print))

    kid.loss_data_train.append(kid.loss)
    kid.eval_data_train.append(kid.evals)

    # Cache the current batch as the monitoring batch if the flag is set.
    if kid.do_batch_monitoring:
        kid.cached_data = kid.sensor.data
        kid.batch_monitoring(kid.cached_data)

    # Initial validation
    if not kid.skip_validation:
        kid.loss, kid.evals = kid.validate()
        kid.on_val_log_step()
        kid.sensor.set_mode("train")
        kid.sensor.setup()


def on_batch_begin(kid):
    pass


def on_epoch_end(kid):
    log.info("Epoch {} finished.".format(kid.epoch))



"""
This module holds classes that standardize evaluation metrics on algorithms.
"""
from __future__ import division

from __future__ import absolute_import
import numpy as np
from sklearn.metrics import roc_auc_score

from .blocks import DataBlock


class EvalBlock(DataBlock):
    """
    Evaluation blocks serve the function to accumulate evaluation metric on
    batches, and summarize the accumulated results on these batches.

    It overrides the `+` operator to accumulate evaluation results. Current
    summary on the accumulated batches is obtainable through the `data`
    property.

    For example, to register an `EvalBlock` for a tensor::

        from akid import backend as A
        from akid import BatchEvalBlock

        a = ... # A is a tensor with a name
        A.register_eval_block(A.get_name(a), BatchEvalBlock)

    It also provide `__str__` method to convert evaluation metrics to a text
    representation. Thus, it may enable a uniform logging interface.
    """
    NAME = "Eval"

    def __str__(self, data=None):
        """
        When taking argument `data`, it converts the `data` to text
        representation, instead of converting the internal data it holds now.
        """
        if data is None:
            data = self.data

        return str(self.data)


class BatchEvalBlock(EvalBlock):
    # TODO: it has a bug. If the last batch is smaller, the weights of its
    # samples are increased, thus leading to wrong accuracy calculation.
    NAME = "BatchEval"

    def __init__(self, v=0):
        self.acc_value = v
        self.num_batches = 0 if v == 0 else 1

    def add(self, v):
        self.acc_value += v
        self.num_batches += 1

    @property
    def data(self):
        return self.acc_value / self.num_batches

    def __str__(self, data=None):
        if data is None:
            data = self.data

        return "{:0.4f}".format(data)


class MultiEvalBlock(EvalBlock):
    """
    The EvalBlock that keeps and accumulates a list of accuracy.
    """
    def __init__(self, v=None):
        self.acc_value = v
        self.num_batches = [0] * len(v) if v is not None else None

    def add(self, v):
        if self.acc_value is None:
            self.acc_value = list(v[:])
            self.num_batches = [1] * len(self.acc_value)
            # Remove possible nan values.
            for i, v in enumerate(self.acc_value):
                if np.isnan(v):
                    self.acc_value[i] = 0
                    self.num_batches[i] = 0
        else:
            for i, _ in enumerate(self.acc_value):
                if not np.isnan(v[i]):
                    # Sometimes, we do not have certain info in a batch. In
                    # this case the value should be passed as nan, and we do
                    # accumulate the results.
                    self.acc_value[i] += v[i]
                    self.num_batches[i] += 1

    @property
    def data(self):
        return [v / self.num_batches[i] for i, v in enumerate(self.acc_value)]

    def __str__(self, data=None):
        if data is None:
            data = self.data

            data_str = ("{:0.4f} " * len(data)).format(*data)
            num_batches_str = ("{} " * len(self.num_batches)).format(*self.num_batches)
            return data_str + num_batches_str
        else:
            return ("{:0.4f} " * len(data)).format(*data)



class AUCEvalBlock(EvalBlock):
    NAME = "AUC"

    def __init__(self):
        self.y_pred = []
        self.y = []

    def add(self, v):
        self.y_pred.append(v[0])
        self.y.append(v[1])

    @property
    def data(self):
        y = np.stack(self.y)
        y_pred = np.stack(self.y_pred)

        assert y.shape[1] == len(self.y[0]), "`np.stack` works wrongly."

        auc = []
        for i in range(y.shape[1]):
            if len(y[:, i]) > 1:
                if not all(y[1:, i] == y[0, i]):
                    auc.append(roc_auc_score(y[:, i], y_pred[:, i], average=None))
        if len(auc) == 0:
            col_auc = None
        else:
            col_auc = np.mean(auc)

        auc = []
        for i in range(y.shape[0]):
            if len(y[i, :]) > 1:
                if not all(y[i, 1:] == y[i, 0]):
                    auc.append(roc_auc_score(y[i, :], y_pred[i, :], average=None))
        if len(auc) == 0:
            row_auc = None
        else:
            row_auc = np.mean(auc)

        return [row_auc, col_auc]

    def __str__(self, data=None):
        if data is None:
            data = self.data

        return "Row AUC: {}, Col AUC {}".format(data[0], data[1])



# Eval blocks for evaluations.
# #######################################################################
tensor_name_to_eval_block_map = {}


def register_eval_block(tensor_name, eval_block):
    """
    For different evaluation metrics, it may take different ways to combine
    the evaluation obtained in batches. This function registers an
    `EvalBlock` to handle the combination for evaluation results obtained
    in the tensor named `tensor_name`.

    If for an evaluation metric tensor that is not registered, `BatchEvalBlock`
    be its eval block by default.
    """
    tensor_name_to_eval_block_map[tensor_name] = eval_block


def get_eval_block(tensor_name):
    try:
        return tensor_name_to_eval_block_map[tensor_name]
    except KeyError:
        return BatchEvalBlock


def reset_eval_block_map():
    global tensor_name_to_eval_block_map
    tensor_name_to_eval_block_map = {}



from __future__ import absolute_import
from six.moves import range
def setup(graph=None):

    from akid import GraphBrain
    from akid.sugar import cnn_block
    from akid import AKID_DATA_PATH
    from akid import Cifar10FeedSource, FeedSensor, Kid
    from akid import MomentumKongFu

    brain = GraphBrain(name="spcnn")

    out_channel_num_list = [64, 128, 256, 512]
    group_size_list = [2, 4, 8, 16]

    for i in range(0, 4):
        brain.attach(cnn_block(
            ksize=[3, 3],
            init_para={
                "name": "uniform",
                "range": 0.005},
            wd={"type": "l2", "scale": 0.005},
            out_channel_num=out_channel_num_list[i],
            pool_size=[2, 2],
            pool_stride=[2, 2],
            activation={"type": "linearize", "group_size": group_size_list[i]},
            keep_prob=0.5,
            bn={"gamma_init": 1, "fix_gamma": True}))

    brain.attach(cnn_block(
        init_para={
            "name": "uniform",
            "range": 0.005},
        wd={"type": "l2", "scale": 0.005},
        out_channel_num=10,
        activation={"type": "softmax"},
        bn={"gamma_init": 1, "fix_gamma": True}))

    # Set up a sensor.
    # #########################################################################
    cifar_source = Cifar10FeedSource(
        name="CIFAR10",
        url='http://www.cs.toronto.edu/~kriz/cifar-10-binary.tar.gz',
        work_dir=AKID_DATA_PATH + '/cifar10',
        use_zca=True,
        num_train=50000,
        num_val=10000)

    sensor = FeedSensor(source_in=cifar_source,
                        batch_size=128,
                        name='data')

    # Summon a survivor.
    # #########################################################################
    survivor = Kid(
        sensor,
        brain,
        MomentumKongFu(base_lr=1.0,
                       momentum=0.5,
                       decay_rate=0.1,
                       decay_epoch_num=25),
        max_steps=60000,
        graph=graph,
    )

    survivor.setup()
    return survivor

kid = setup()
kid.practice()



"""
This module provides systems of different topology to compose `Block`s to
create more complex blocks. A system does not concern which type of block it
holds, but only concerns the mathematical topology how they connect.

Atrributes defined in a finer granularity overrides the same attributes defined
in a system. E.g., if `do_summary` is defined as `False` in a block, and the
block is attached to a system with `do_summary` valued `True`, the `False` will
take priority.
"""
from __future__ import absolute_import
import copy

from .blocks import GenerativeBlock
from .. import backend as A
from .interface_blocks import UpdateBlock


class System(GenerativeBlock, UpdateBlock):
    """
    A top level class to model a system that is purposeless. It means this
    system does not serve a clear purpose, but an aggregation of blocks.

    If you want to make it of purpose, or in another word, enforce semantics on
    it, you could combine with other classes by multi-inheritance and override
    `attach` method to add a gate keeper. For instance, a `Brain` should only
    contain `ProcessingLayer` since its purpose is to processing data. So
    `Brain` could also be taken as a `ProcessingLayer` layer and it is also a
    system.

    It enforces an interface `data` to provide the data propagated in this
    system.

    A system has direct access to the outputs of the blocks it contains, but it
    only should use them when an action that cannot be accomplished without
    information from more than one blocks.
    """

    def __init__(self, do_stat_on_norm=False, **kwargs):
        super(System, self).__init__(**kwargs)

        self.do_stat_on_norm = do_stat_on_norm

        self.blocks = []

        # Quick lookup table to get the idx of a block. Since the blocks are
        # added in a topological order normally, this table provides quick
        # topological order lookup.
        self.block_name_to_idx = {}

    def get_copy(self):
        self_copy = super(System, self).get_copy()

        self_copy.blocks = []
        for b in self.blocks:
            self_copy.blocks.append(b.get_copy())

        return self_copy

    def get_clone(self):
        clone = super(System, self).get_copy()

        clone.blocks = []
        for b in self.blocks:
            clone.blocks.append(b.get_clone())

        return clone

    def set_shadow(self):
        super(System, self).set_shadow()
        for b in self.blocks:
            b.set_shadow()

    def get_last_layer_name(self):
        """
        Return the name of the last layer that is attached to the system. It is
        useful for getting the name and feed it to another layer as its input.
        """
        return self.blocks[-1].name

    def get_layer_by_name(self, name):
        """
        Get any layers in the system by its name. None if the layer is not
        found.
        """
        for b in self.blocks:
            if b.name is name:
                return b

        raise Exception("Layer {} is not found".format(name))

    @property
    def is_empty(self):
        if self.blocks:
            return False

        return True

    @property
    def data(self):
        """
        Since it only has one output, override method to a property.
        """
        return self._data

    def attach(self, block_in):
        # Set flags value to the value of this system if they are not
        # individually set.
        for f in self.__class__._flags:
            if getattr(block_in, f) is None:
                block_in.set_flag(f, getattr(self, f))

        self.blocks.append(block_in)

        self.block_name_to_idx[block_in.name] = len(self.blocks) - 1

    def set_do_summary_flag(self, v):
        super(System, self).set_do_summary_flag(v)
        for b in self.blocks:
            b.do_summary = v

    def set_flag(self, flag_name, v):
        """
        Set a flag that all blocks in the system all have recursively.
        """
        if flag_name in self.__class__._flags:
            setattr(self, flag_name, v)
        else:
            raise ValueError("{} does not have flag {}".format(self.__class__, flag_name))

        for b in self.blocks:
            b.set_flag(flag_name, v)


class SequentialSystem(System):
    """
    A system that links blocks one by one sequentially, like the linked list
    structure in Computer Science Data Structure.
    """
    def _setup(self):
        for b in self.blocks:
            self.log("Setting up block {}.".format(b.name))
            b.setup()

    def _forward(self, data_in):
        """
        Link the blocks linearly together. It takes exact one argument and
        apply processing blocks to it one by one, and return the final
        processed data.

        A `SequentialSystem` could be used standalone. However, another typical use
        of `SequentialSystem` is inherit it to make something more complex, such as
        creating a `Brain` sequentially linking layers together. In that case,
        `_forward` and `backward` would be overrided.
        """
        data = data_in
        try:
            shape = A.get_shape(data)
        except ValueError:
            shape = None

        if not self.done_first_pass:
            self.log("Doing forward propagation.")
            self.log("System input shape: {}".format(shape))

        previous_data_name = 'system_in'
        for l in self.blocks:

            data_out = l.forward(data)

            if not self.done_first_pass:
                name = A.get_name(data)
                p_name = name if name else previous_data_name
                name = A.get_name(data_out)
                n_name = name if name else l.name
                self.log("Connected: {} -> {}".format(p_name, n_name))
                previous_data_name = l.name

                if shape:
                    self.log("Top shape: {}".format(A.get_shape(data_out)))
            data = l.data

        self._data = data

        return data_out

    def _on_update(self):
        l = self.blocks[0]
        # The Riemannian metric is None for the first layer
        l._on_update(None)

        l_prev = l
        for l in self.blocks[1:]:
            l._on_update(l_prev.K)
            l_prev = l


class GraphSystem(SequentialSystem):
    """
    A system that is capable to handle arbitrary graph style connections
    between blocks.

    It is supposed to contain `ProcessingLayer` that has a `inputs` attributes
    to hold interconnection information between layers.  If `inputs` is None,
    it means this layer is supposed to take all the outputs of previous layer
    (depending on the actually topology of the system this block is in) as its
    input data. A tensor is the input if the previous layer only has one
    output, otherwise, a list of tensor would be the input. The first block of
    the system is special, see docstring of `Brain`. If not None, a list of
    tuples should be passed in. For example, the input of this layer is
    supposed to be outputs of layer "conv_1" and "conv2", then a list of
    [{"name": "conv1", "idxs": [0]}, {"name": "conv2", "idxs": [0]}] should be
    passed in. Output of any blocks are a tuple (if there are multiple
    outputs). The list of indices means the indices of outputs of that layer to
    use.

    It also supports directly pass tensor in.

    NOTE: no matter how the inputs are specified (by `inputs` or not), in the
    `_forward` method of a block, inputs feeds to a block is a tensor (if there
    is only one input) or a list of tensors (if there are multiple inputs.)
    """
    def _forward(self, data_in, injected_data=None, start_block=None, end_block=None):
        """
        Method overrode to handle arbitrary layer interconnections.

        It also supports partial propagation by specifying the names of the
        start and the end blocks. Note that in this case, `data_in` should be
        given in the shape as that of the input of `start_block` block.

        Args:
            data_in: list of tensors
            injected_data: dict
                A dict consists of injected data to replace the inputs that are
                needed in partial propagation.
            start_block: str
                The name of the start block for partial propagation.
            end_block: str
                The name of the end block for partial propagation. If None, and
                `start_block` is given, the system do forward propagation until
                the end.
        """
        # Normalize input to a list for convenience even if there is only one
        # input.
        t = type(data_in)
        data = data_in if t is list or t is tuple else [data_in]

        if not self.done_first_pass:
            self.log("System input shape: {}".format(
                [A.get_shape(d) for d in data]))

        previous_data_name = "system_in"

        # Check if we do partial propagation.
        if start_block is None:
            start_block = self.blocks[0].name

        # Find the start_block and start from it.
        for start_idx,l in enumerate(self.blocks):
            if l.name == start_block:
                break

        if end_block is None:
            end_block = self.blocks[-1].name

        # Find the end block if given.
        for end_idx in range(1, len(self.blocks)+1):
            if self.blocks[-end_idx].name == end_block:
                end_idx = len(self.blocks) - end_idx + 1
                break

        if start_idx > end_idx:
            raise ValueError("Start idx should be larger than the end idx.")

        for i, l in enumerate(self.blocks[start_idx:end_idx]):

            inputs = None
            if l.inputs:
                # Find inputs in the system to current block.
                inputs = []
                for input in l.inputs:
                    if A.is_tensor(input):
                        inputs.append(input)
                        continue

                    input_num = len(inputs)

                    # Pick the first output tensor, if not specifying idxs
                    if "idxs" not in input:
                        input["idxs"] = [0]

                    # First check whether the input is from the system input.
                    if input["name"] == "system_in":
                        for i in input["idxs"]:
                            inputs.append(data_in[i])
                    # Then look through outputs of setup layers.
                    else:
                        block_idx = self.block_name_to_idx[input["name"]]
                        b = self.blocks[block_idx]
                        data = b.data

                        if block_idx < start_idx:
                            # The input should be fed in from system_in.
                            data = injected_data[input["name"]]

                        # If a layer has only one output, directly put
                        # that data in the input since otherwise, this
                        # layer won't be listed at all.
                        if type(data) is not list:
                            inputs.append(data)
                        else:
                            for i in input["idxs"]:
                                inputs.append(data[i])

                    if "idxs" not in input:
                        input_inc = 1
                    else:
                        input_inc = len(input["idxs"])
                    if len(inputs) != input_num + input_inc:
                        raise Exception("{} is not found. You perhaps misspell"
                                        " the layer name.".format(
                                            input["name"]))

                if len(inputs) is 1:
                    l.forward(inputs[0])
                else:
                    l.forward(inputs)
            else:
                # By default, we only pass the first tensor to the new layer.
                l.forward(data[0])

            if not self.done_first_pass:
                # Logging
                dtype = type(data)
                if inputs:
                    in_name = [A.get_name(i) for i in inputs]
                    if None in in_name:
                        # No name is created for the inputs, thus, we use their
                        # shape instead.
                        in_name = [str(i.shape) for i in inputs]
                else:
                    in_name = A.get_name(data[0])

                if in_name is None:
                    in_name = previous_data_name

                if l.data is not None:
                    dtype = type(l.data)
                    if dtype is not tuple and dtype is not list:
                        n = A.get_name(l.data)
                        out_name = n if n else l.name
                    else:
                        out_name = [A.get_name(d) for d in l.data]
                        if None in out_name:
                            out_name = l.name

                    self.log("Connected: {} -> {}".format(in_name, out_name))
                    self.log("Top shape: {}".format(
                        A.get_shape(l.data) if dtype is not tuple and dtype is not list
                        else [A.get_shape(d) for d in l.data]))
                else:
                    self.log("Inputs: {}. No outputs.".format(in_name))

            dtype = type(l.data)
            if l.data is not None:
                data = l.data if dtype is tuple or dtype is list else [l.data]
            else:
                data = None

            previous_data_name = l.name

        self._data = data

        return data


class SequentialGSystem(GraphSystem):
    """
    A `System` that supports arbitrary connectivity forward propagation, but
    backward propagation only supports sequentially connected blocks. Roughly,
    it is a generative system that supports `GraphSystem` way of specifying
    labels.
    """
    def backward(self, X_in):
        """
        Call `backward` in each linked block in reverse order.

        NOTE: The sequentially linked version is written first given a topology
        order may need to be inferred if we want to do arbitrary connectivity.
        """
        # Assume only one input
        if type(X_in) is list:
            X_in = X_in[0]

        data = X_in
        try:
            shape = A.get_shape(data)
        except ValueError:
            shape = None

        self.log("Doing backward propagation.")
        self.log("System input shape: {}".format(shape))
        for l in reversed(self.blocks):
            l.backward(data)
            self.log("Connected: {} -> {}".format(data.name,
                                                  l.data_g.name))
            if shape:
                self.log("Top shape: {}".format(A.get_shape(l.data)))
            data = l.data_g

        self._data_g = data

        return self._data_g
